# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_config import cfg
from oslo_messaging import exceptions
from oslo_utils import uuidutils
import testtools

from neutron.tests import base

from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.cfg_agent import cfg_agent


_uuid = uuidutils.generate_uuid
HOSTNAME = 'myhost'
FAKE_ID = _uuid()


def prepare_router_data(enable_snat=None, num_internal_ports=1):
    router_id = _uuid()
    ex_gw_port = {'id': _uuid(),
                  'network_id': _uuid(),
                  'fixed_ips': [{'ip_address': '19.4.4.4',
                                 'subnet_id': _uuid()}],
                  'subnet': {'cidr': '19.4.4.0/24',
                             'gateway_ip': '19.4.4.1'}}
    int_ports = []
    for i in range(num_internal_ports):
        int_ports.append({'id': _uuid(),
                          'network_id': _uuid(),
                          'admin_state_up': True,
                          'fixed_ips': [{'ip_address': '35.4.%s.4' % i,
                                         'subnet_id': _uuid()}],
                          'mac_address': 'ca:fe:de:ad:be:ef',
                          'subnet': {'cidr': '35.4.%s.0/24' % i,
                                     'gateway_ip': '35.4.%s.1' % i}})
    hosting_device = {'id': _uuid(),
                      'host_type': 'VM',
                      'ip_address': '20.0.0.5',
                      'port': '23'}

    router = {
        'id': router_id,
        bc.constants.INTERFACE_KEY: int_ports,
        'routes': [],
        'gw_port': ex_gw_port,
        'hosting_device': hosting_device}
    if enable_snat is not None:
        router['enable_snat'] = enable_snat
    return router, int_ports


class TestCiscoCfgAgentWithStateReporting(base.BaseTestCase):

    def setUp(self):
        self.conf = cfg.ConfigOpts()
        bc.config.register_agent_state_opts_helper(cfg.CONF)
        self.conf.register_opts(bc.core_opts)
        self.conf.register_opts(cfg_agent.OPTS, "cfg_agent")
        cfg.CONF.set_override('report_interval', 0, 'AGENT')
        super(TestCiscoCfgAgentWithStateReporting, self).setUp()
        self.devmgr_plugin_api_cls_p = mock.patch(
            'networking_cisco.plugins.cisco.cfg_agent.cfg_agent.'
            'CiscoDeviceManagementApi')
        devmgr_plugin_api_cls = self.devmgr_plugin_api_cls_p.start()
        self.devmgr_plugin_api = mock.Mock()
        devmgr_plugin_api_cls.return_value = self.devmgr_plugin_api
        self.devmgr_plugin_api.register_for_duty.return_value = True

        self.plugin_reportstate_api_cls_p = mock.patch(
            'neutron.agent.rpc.PluginReportStateAPI')
        plugin_reportstate_api_cls = self.plugin_reportstate_api_cls_p.start()
        self.plugin_reportstate_api = mock.Mock()
        plugin_reportstate_api_cls.return_value = self.plugin_reportstate_api

        self.looping_call_p = mock.patch(
            'oslo_service.loopingcall.FixedIntervalLoopingCall')
        self.looping_call_p.start()

        mock.patch('neutron.common.rpc.create_connection').start()

    def test_agent_registration_success(self):
        agent = cfg_agent.CiscoCfgAgentWithStateReport(HOSTNAME, self.conf)
        self.assertTrue(agent.devmgr_rpc.register_for_duty(agent.context))

    def test_agent_registration_success_after_2_tries(self):
        self.devmgr_plugin_api.register_for_duty = mock.Mock(
            side_effect=[False, False, True])
        cfg_agent.REGISTRATION_RETRY_DELAY = 0.01
        agent = cfg_agent.CiscoCfgAgentWithStateReport(HOSTNAME, self.conf)
        self.assertEqual(3, agent.devmgr_rpc.register_for_duty.call_count)

    def test_agent_registration_fail_always(self):
        self.devmgr_plugin_api.register_for_duty = mock.Mock(
            return_value=False)
        cfg_agent.REGISTRATION_RETRY_DELAY = 0.01
        cfg_agent.MAX_REGISTRATION_ATTEMPTS = 3
        with testtools.ExpectedException(SystemExit):
            cfg_agent.CiscoCfgAgentWithStateReport(HOSTNAME, self.conf)

    def test_agent_registration_no_device_mgr(self):
        self.devmgr_plugin_api.register_for_duty = mock.Mock(
            return_value=None)
        cfg_agent.REGISTRATION_RETRY_DELAY = 0.01
        cfg_agent.MAX_REGISTRATION_ATTEMPTS = 3
        with testtools.ExpectedException(SystemExit):
            cfg_agent.CiscoCfgAgentWithStateReport(HOSTNAME, self.conf)

    def test_assigned_hosting_devices_monitored_from_start(self):
        agent = cfg_agent.CiscoCfgAgentWithStateReport(HOSTNAME, self.conf)
        fake_hds = [{'id': 'fake_id1'}, {'id': 'fake_id2'}]
        with mock.patch.object(agent.devmgr_rpc,
                               'get_hosting_devices_for_agent',
                               return_value=fake_hds), \
                mock.patch.object(agent, '_dev_status') as ds_mock:
            agent.after_start()
            ds_mock.backlog_hosting_devices.assert_called_once_with(fake_hds)

    def test_assigned_hosting_devices_monitored_from_start_retry(self):
        agent = cfg_agent.CiscoCfgAgentWithStateReport(HOSTNAME, self.conf)
        fake_hds = [{'id': 'fake_id1'}, {'id': 'fake_id2'}]
        results = [exceptions.MessagingException(),
                   exceptions.MessagingException(), fake_hds]
        with mock.patch.object(agent.devmgr_rpc,
                               'get_hosting_devices_for_agent',
                               side_effect=results), \
                mock.patch.object(agent, '_dev_status') as ds_mock:
            agent.after_start()
            ds_mock.backlog_hosting_devices.assert_called_once_with(fake_hds)

    def test_no_hosting_devices_monitored_from_start_if_rpc_fail(self):
        agent = cfg_agent.CiscoCfgAgentWithStateReport(HOSTNAME, self.conf)
        with mock.patch.object(agent.devmgr_rpc,
                               'get_hosting_devices_for_agent',
                               side_effect=exceptions.MessagingException), \
                mock.patch.object(agent, '_dev_status') as ds_mock:
            agent.after_start()
            ds_mock.backlog_hosting_devices.assert_not_called()

    def test_report_state(self):
        agent = cfg_agent.CiscoCfgAgentWithStateReport(HOSTNAME, self.conf)
        # Set keepalive iteration to just before the reporting iteration
        agent.keepalive_iteration = self.conf.cfg_agent.report_iteration - 1
        agent._report_state()
        self.assertIn('total routers', agent.agent_state['configurations'])
        self.assertEqual(0,
                         agent.agent_state['configurations']['total routers'])

    def test_report_state_report_iteration_check_full_report(self):
        agent = cfg_agent.CiscoCfgAgentWithStateReport(HOSTNAME, self.conf)
        # Set keepalive iteration to just before the reporting iteration
        agent.keepalive_iteration = self.conf.cfg_agent.report_iteration - 1
        agent._report_state()
        self.assertIn('total routers', agent.agent_state['configurations'])
        self.assertEqual(0, agent.agent_state[
            'configurations']['total routers'])
        self.assertEqual(0, agent.keepalive_iteration)

    def test_report_state_report_iteration_check_partial_report(self):
        agent = cfg_agent.CiscoCfgAgentWithStateReport(HOSTNAME, self.conf)
        # Retain original keepalive iteration
        agent.keepalive_iteration = self.conf.cfg_agent.report_iteration
        agent._report_state()
        self.assertNotIn('configurations', agent.agent_state)
        self.assertEqual((self.conf.cfg_agent.report_iteration + 1),
                         agent.keepalive_iteration)

    @mock.patch('networking_cisco.plugins.cisco.cfg_agent.'
                'cfg_agent.CiscoCfgAgentWithStateReport._agent_registration')
    def test_report_state_attribute_error(self, agent_registration):
        cfg.CONF.set_override('report_interval', 1, 'AGENT')
        self.plugin_reportstate_api.report_state.side_effect = AttributeError
        agent = cfg_agent.CiscoCfgAgentWithStateReport(HOSTNAME, self.conf)
        agent.heartbeat = mock.Mock()
        agent.send_agent_report(None, None)
        self.assertTrue(agent.heartbeat.stop.called)

    def test_get_hosting_device_configuration(self):
        routing_service_helper_mock = mock.MagicMock()
        routing_service_helper_mock.driver_manager = mock.MagicMock()
        drv_mgr = routing_service_helper_mock.driver_manager
        drv = drv_mgr.get_driver_for_hosting_device.return_value
        fake_running_config = 'a fake running config'
        drv.get_configuration = mock.MagicMock(
            return_value=fake_running_config)
        hd_id = 'a_hd_id'
        payload = {'hosting_device_id': hd_id}
        agent = cfg_agent.CiscoCfgAgentWithStateReport(HOSTNAME, self.conf)
        agent.routing_service_helper = routing_service_helper_mock
        res = agent.get_hosting_device_configuration(mock.MagicMock(), payload)
        self.assertEqual(fake_running_config, res)
        drv.get_configuration.assert_called_once_with()

    def test_get_hosting_device_configuration_no_hosting_device(self):
        routing_service_helper_mock = mock.MagicMock()
        routing_service_helper_mock.driver_manager = mock.MagicMock()
        drv_mgr = routing_service_helper_mock.driver_manager
        drv = drv_mgr.get_driver_for_hosting_device.return_value
        fake_running_config = 'a fake running config'
        drv.get_configuration = mock.MagicMock(
            return_value=fake_running_config)
        hd_id = None
        payload = {'hosting_device_id': hd_id}
        agent = cfg_agent.CiscoCfgAgentWithStateReport(HOSTNAME, self.conf)
        agent.routing_service_helper = routing_service_helper_mock
        res = agent.get_hosting_device_configuration(mock.MagicMock(), payload)
        self.assertIsNone(res)
        drv.get_configuration.assert_not_called()

    def test_get_hosting_device_configuration_no_svc_helper(self):
        routing_service_helper_mock = mock.MagicMock()
        routing_service_helper_mock.driver_manager = mock.MagicMock()
        drv_mgr = routing_service_helper_mock.driver_manager
        drv = drv_mgr.get_driver_for_hosting_device.return_value
        fake_running_config = 'a fake running config'
        drv.get_configuration = mock.MagicMock(
            return_value=fake_running_config)
        hd_id = 'a_hd_id'
        payload = {'hosting_device_id': hd_id}
        agent = cfg_agent.CiscoCfgAgentWithStateReport(HOSTNAME, self.conf)
        agent.routing_service_helper = None
        res = agent.get_hosting_device_configuration(mock.MagicMock(), payload)
        self.assertIsNone(res)
        drv.get_configuration.assert_not_called()

    def test_plugin_notified_about_revived_hosting_devices_heartbeat_on(self):
        self.conf.set_override('enable_heartbeat', True, 'cfg_agent')
        agent = cfg_agent.CiscoCfgAgentWithStateReport(HOSTNAME, self.conf)
        _dev_status_mock = mock.MagicMock()
        ids_revived_hds = ['fake_hd_id1', 'fake_hd_id2']
        _dev_status_mock.check_backlogged_hosting_devices.return_value = {
            'reachable': [], 'dead': [], 'revived': ids_revived_hds}
        with mock.patch.object(agent, '_dev_status', _dev_status_mock),\
                mock.patch.object(agent, 'process_services') as p_s_mock,\
                mock.patch.object(agent, 'devmgr_rpc') as d_r_mock:
            ctx = mock.MagicMock()
            agent._process_backlogged_hosting_devices(ctx)
        p_s_mock.assert_called_with(device_ids=ids_revived_hds)
        d_r_mock.report_revived_hosting_devices.assert_called_with(
            ctx, hd_ids=ids_revived_hds)

    def test_plugin_not_notified_about_revived_hosting_devices_heartbeat_off(
            self):
        self.conf.set_override('enable_heartbeat', False, 'cfg_agent')
        agent = cfg_agent.CiscoCfgAgentWithStateReport(HOSTNAME, self.conf)
        _dev_status_mock = mock.MagicMock()
        ids_revived_hds = ['fake_hd_id1', 'fake_hd_id2']
        _dev_status_mock.check_backlogged_hosting_devices.return_value = {
            'reachable': [], 'dead': [], 'revived': ids_revived_hds}
        with mock.patch.object(agent, '_dev_status', _dev_status_mock),\
                mock.patch.object(agent, 'process_services') as p_s_mock,\
                mock.patch.object(agent, 'devmgr_rpc') as d_r_mock:
            ctx = mock.MagicMock()
            agent._process_backlogged_hosting_devices(ctx)
        self.assertEqual(0, p_s_mock.call_count)
        self.assertEqual(0, d_r_mock.report_revived_hosting_devices.call_count)



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import webob.exc

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import uuidutils
import six

from networking_cisco._i18n import _
from networking_cisco import backwards_compatibility as bc

LOG = logging.getLogger(__name__)


def get_specific_config(prefix):
    """Retrieve config based on the format [<prefix>:<value>].

    returns: a dict, {<UUID>: {<key1>:<value1>, <key2>:<value2>, ...}}
    """
    conf_dict = {}
    multi_parser = cfg.MultiConfigParser()
    read_ok = multi_parser.read(cfg.CONF.config_file)
    if len(read_ok) != len(cfg.CONF.config_file):
        raise cfg.Error(_("Some config files were not parsed properly"))

    for parsed_file in multi_parser.parsed:
        for parsed_item in parsed_file.keys():
            p_i = parsed_item.lower()
            if p_i.startswith(prefix):
                section_type, uuid = p_i.split(':')
                if section_type == prefix:
                    conf_dict[uuid] = {k: v[0] for (k, v) in parsed_file[
                        parsed_item].items()}
    return conf_dict


def verify_resource_dict(res_dict, is_create, attr_info):
    """Verifies required attributes are in resource dictionary, res_dict.

    Also checking that an attribute is only specified if it is allowed
    for the given operation (create/update).

    Attribute with default values are considered to be optional.

    This function contains code taken from function 'prepare_request_body' in
    attributes.py.
    """
    if ((bc.NEUTRON_VERSION >= bc.NEUTRON_NEWTON_VERSION) and 'tenant_id'
            in res_dict):
        res_dict['project_id'] = res_dict['tenant_id']
    if is_create:  # POST
        for attr, attr_vals in six.iteritems(attr_info):
            if attr_vals['allow_post']:
                if 'default' not in attr_vals and attr not in res_dict:
                    msg = _("Failed to parse request. Required attribute '%s' "
                            "not specified") % attr
                    raise webob.exc.HTTPBadRequest(msg)
                res_dict[attr] = res_dict.get(attr, attr_vals.get('default'))
            else:
                if attr in res_dict:
                    msg = _("Attribute '%s' not allowed in POST") % attr
                    raise webob.exc.HTTPBadRequest(msg)
    else:  # PUT
        for attr, attr_vals in six.iteritems(attr_info):
            if attr in res_dict and not attr_vals['allow_put']:
                msg = _("Cannot update read-only attribute %s") % attr
                raise webob.exc.HTTPBadRequest(msg)

    for attr, attr_vals in six.iteritems(attr_info):
        if (attr not in res_dict or
                res_dict[attr] is bc.constants.ATTR_NOT_SPECIFIED):
            continue
        # Convert values if necessary
        if 'convert_to' in attr_vals:
            res_dict[attr] = attr_vals['convert_to'](res_dict[attr])
        # Check that configured values are correct
        if 'validate' not in attr_vals:
            continue
        for rule in attr_vals['validate']:
            _ensure_format(rule, attr, res_dict)
            res = bc.validators[rule](res_dict[attr],
                                      attr_vals['validate'][rule])
            if res:
                msg_dict = dict(attr=attr, reason=res)
                msg = (_("Invalid input for %(attr)s. Reason: %(reason)s.") %
                       msg_dict)
                raise webob.exc.HTTPBadRequest(msg)
    return res_dict


def uuidify(val):
    """Takes an integer and transforms it to a UUID format.

    returns: UUID formatted version of input.
    """
    if uuidutils.is_uuid_like(val):
        return val
    else:
        try:
            int_val = int(val, 16)
        except ValueError:
            with excutils.save_and_reraise_exception():
                LOG.error("Invalid UUID format %s. Please provide an "
                          "integer in decimal (0-9) or hex (0-9a-e) "
                          "format", val)
        res = str(int_val)
        num = 12 - len(res)
        return "00000000-0000-0000-0000-" + "0" * num + res


def _ensure_format(rule, attribute, res_dict):
    """Verifies that attribute in res_dict is properly formatted.

    Since, in the .ini-files, lists are specified as ':' separated text and
    UUID values can be plain integers we need to transform any such values
    into proper format. Empty strings are converted to None if validator
    specifies that None value is accepted.
    """
    if rule == 'type:uuid' or (rule == 'type:uuid_or_none' and
                               res_dict[attribute]):
        res_dict[attribute] = uuidify(res_dict[attribute])
    elif rule == 'type:uuid_list':
        if not res_dict[attribute]:
            res_dict[attribute] = []
        else:
            temp_list = res_dict[attribute].split(':')
            res_dict[attribute] = []
            for item in temp_list:
                res_dict[attribute].append = uuidify(item)
    elif rule == 'type:string_or_none' and res_dict[attribute] == "":
        res_dict[attribute] = None


def obtain_hosting_device_credentials_from_config():
    """Obtains credentials from config file and stores them in memory.
    To be called before hosting device templates defined in the config file
    are created.
    """
    cred_dict = get_specific_config('cisco_hosting_device_credential')
    attr_info = {
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:string': None}, 'is_visible': True,
                 'default': ''},
        'description': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:string': None},
                        'is_visible': True, 'default': ''},
        'user_name': {'allow_post': True, 'allow_put': True,
                      'validate': {'type:string': None},
                      'is_visible': True, 'default': ''},
        'password': {'allow_post': True, 'allow_put': True,
                     'validate': {'type:string': None},
                     'is_visible': True, 'default': ''},
        'type': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:string': None}, 'is_visible': True,
                 'default': ''}}
    credentials = {}
    for cred_uuid, kv_dict in cred_dict.items():
        # ensure cred_uuid is properly formatted
        cred_uuid = uuidify(cred_uuid)
        verify_resource_dict(kv_dict, True, attr_info)
        credentials[cred_uuid] = kv_dict
    return credentials



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


from six.moves import queue
import time

from oslo_serialization import jsonutils

from networking_cisco.apps.saf.agent import detect_uplink as uplink_det
from networking_cisco.apps.saf.agent.topo_disc import topo_disc
from networking_cisco.apps.saf.agent.vdp import ovs_vdp
from networking_cisco.apps.saf.common import constants
from networking_cisco.apps.saf.common import dfa_logger as logging
from networking_cisco.apps.saf.common import dfa_sys_lib as sys_utils
from networking_cisco.apps.saf.common import rpc
from networking_cisco.apps.saf.common import utils

LOG = logging.getLogger(__name__)


class VdpMsgPriQue(object):

    """VDP Message Queue. """

    def __init__(self):
        self._queue = queue.PriorityQueue()

    def enqueue(self, priority, msg):
        msg_tupl = (priority, msg)
        self._queue.put(msg_tupl)

    def dequeue(self):
        msg_tupl = self._queue.get()
        return msg_tupl[0], msg_tupl[1]

    def dequeue_nonblock(self):
        msg_tupl = self._queue.get_nowait()
        return msg_tupl[0], msg_tupl[1]

    def is_not_empty(self):
        return not self._queue.empty()


class VdpQueMsg(object):

    """Construct VDP Message. """

    def __init__(self, msg_type, port_uuid=None, vm_mac=None, oui=None,
                 net_uuid=None, segmentation_id=None, status=None,
                 vm_bulk_list=None, phy_uplink=None, br_int=None, br_ex=None,
                 root_helper=None):
        self.msg_dict = {}
        self.msg_type = msg_type
        if msg_type == constants.VM_MSG_TYPE:
            self.construct_vm_msg(port_uuid, vm_mac, net_uuid,
                                  segmentation_id, status, oui, phy_uplink)
        elif msg_type == constants.UPLINK_MSG_TYPE:
            self.construct_uplink_msg(status, phy_uplink, br_int, br_ex,
                                      root_helper)
        elif msg_type == constants.VM_BULK_SYNC_MSG_TYPE:
            self.construct_vm_bulk_sync_msg(vm_bulk_list, phy_uplink)

    def construct_vm_msg(self, port_uuid, vm_mac, net_uuid,
                         segmentation_id, status, oui, phy_uplink):
        self.msg_dict['port_uuid'] = port_uuid
        self.msg_dict['vm_mac'] = vm_mac
        self.msg_dict['net_uuid'] = net_uuid
        self.msg_dict['segmentation_id'] = segmentation_id
        self.msg_dict['status'] = status
        self.msg_dict['oui'] = oui
        self.msg_dict['phy_uplink'] = phy_uplink

    def construct_vm_bulk_sync_msg(self, vm_bulk_list, phy_uplink):
        self.msg_dict['phy_uplink'] = phy_uplink
        self.msg_dict['vm_bulk_list'] = vm_bulk_list

    def construct_uplink_msg(self, status, phy_uplink, br_int, br_ex,
                             root_helper):
        self.msg_dict['status'] = status
        self.msg_dict['phy_uplink'] = phy_uplink
        self.msg_dict['br_int'] = br_int
        self.msg_dict['br_ex'] = br_ex
        self.msg_dict['root_helper'] = root_helper

    def get_oui(self):
        return self.msg_dict['oui']

    def get_uplink(self):
        return self.msg_dict['phy_uplink']

    def get_mac(self):
        return self.msg_dict['vm_mac']

    def get_status(self):
        return self.msg_dict['status']

    def get_segmentation_id(self):
        return self.msg_dict['segmentation_id']

    def get_net_uuid(self):
        return self.msg_dict['net_uuid']

    def get_port_uuid(self):
        return self.msg_dict['port_uuid']

    def get_integ_br(self):
        return self.msg_dict['br_int']

    def get_ext_br(self):
        return self.msg_dict['br_ex']

    def get_root_helper(self):
        return self.msg_dict['root_helper']

    def set_uplink(self, uplink):
        self.msg_dict['phy_uplink'] = uplink


class VdpMgr(object):

    """Responsible for Handling VM/Uplink requests. """

    def __init__(self, config_dict, rpc_client, hostname):
        self.br_integ = config_dict.get('integration_bridge')
        self.br_ex = config_dict.get('external_bridge')
        self.root_helper = config_dict.get('root_helper')
        self.host_id = config_dict.get('host_id')
        self.node_list = config_dict['node_list']
        self.node_uplink_list = config_dict['node_uplink_list']
        # Check for error?? fixme(padkrish)
        self.que = VdpMsgPriQue()
        self.err_que = VdpMsgPriQue()
        self.phy_uplink = None
        self.veth_intf = None
        self.restart_uplink_called = False
        self.ovs_vdp_obj_dict = {}
        self.rpc_clnt = rpc_client
        self.host_name = hostname
        self.uplink_det_compl = False
        self.process_uplink_ongoing = False
        self.uplink_down_cnt = 0
        self.is_os_run = False
        self.static_uplink = False
        self.static_uplink_port = None
        self.static_uplink_first = True
        self.bulk_vm_rcvd_flag = False
        self.bulk_vm_check_cnt = 0
        self.vdp_mgr_lock = utils.lock()
        self.read_static_uplink()
        self.start()
        self.topo_disc = topo_disc.TopoDisc(self.topo_disc_cb,
                                            self.root_helper)

    def read_static_uplink(self):
        """Read the static uplink from file, if given."""
        if self.node_list is None or self.node_uplink_list is None:
            return
        for node, port in zip(self.node_list.split(','),
                              self.node_uplink_list.split(',')):
            if node.strip() == self.host_name:
                self.static_uplink = True
                self.static_uplink_port = port.strip()
                return

    def topo_disc_cb(self, intf, topo_disc_obj):
        return self.save_topo_disc_params(intf, topo_disc_obj)

    def update_vm_result(self, port_uuid, result, lvid=None,
                         vdp_vlan=None, fail_reason=None):
        context = {'agent': self.host_id}
        if lvid is None or vdp_vlan is None:
            args = jsonutils.dumps({'port_uuid': port_uuid, 'result': result,
                                    'fail_reason': fail_reason})
        else:
            args = jsonutils.dumps({'port_uuid': port_uuid, 'local_vlan': lvid,
                                    'vdp_vlan': vdp_vlan, 'result': result,
                                    'fail_reason': fail_reason})
        msg = self.rpc_clnt.make_msg('update_vm_result', context, msg=args)
        try:
            resp = self.rpc_clnt.call(msg)
            return resp
        except rpc.MessagingTimeout:
            LOG.error("RPC timeout: Failed to update VM result on the"
                      " server")

    def vdp_vlan_change_cb(self, port_uuid, lvid, vdp_vlan, fail_reason):
        """Callback function for updating the VDP VLAN in DB. """
        LOG.info("Vlan change CB lvid %(lvid)s VDP %(vdp)s",
                 {'lvid': lvid, 'vdp': vdp_vlan})
        self.update_vm_result(port_uuid, constants.RESULT_SUCCESS,
                              lvid=lvid, vdp_vlan=vdp_vlan,
                              fail_reason=fail_reason)

    def process_vm_event(self, msg, phy_uplink):
        LOG.info("In processing VM Event status %(status)s for MAC "
                 "%(mac)s UUID %(uuid)s oui %(oui)s",
                 {'status': msg.get_status(), 'mac': msg.get_mac(),
                  'uuid': msg.get_port_uuid(), 'oui': msg.get_oui()})
        time.sleep(10)
        if msg.get_status() == 'up':
            res_fail = constants.CREATE_FAIL
        else:
            res_fail = constants.DELETE_FAIL
        if (not self.uplink_det_compl or
                phy_uplink not in self.ovs_vdp_obj_dict):
            LOG.error("Uplink Port Event not received yet")
            self.update_vm_result(msg.get_port_uuid(), res_fail)
            return
        ovs_vdp_obj = self.ovs_vdp_obj_dict[phy_uplink]
        port_event_reply = ovs_vdp_obj.send_vdp_port_event(
            msg.get_port_uuid(), msg.get_mac(), msg.get_net_uuid(),
            msg.get_segmentation_id(), msg.get_status(), msg.get_oui())
        if not port_event_reply.get('result'):
            LOG.error("Error in VDP port event, Err Queue enq")
            self.update_vm_result(
                msg.get_port_uuid(), res_fail,
                fail_reason=port_event_reply.get('fail_reason'))
        else:
            LOG.info("Success in VDP port event")
            lvid, vdp_vlan = ovs_vdp_obj.get_lvid_vdp_vlan(msg.get_net_uuid(),
                                                           msg.get_port_uuid())
            self.update_vm_result(
                msg.get_port_uuid(), constants.RESULT_SUCCESS,
                lvid=lvid, vdp_vlan=vdp_vlan,
                fail_reason=port_event_reply.get('fail_reason'))

    def process_bulk_vm_event(self, msg, phy_uplink):
        """Process the VM bulk event usually after a restart. """
        LOG.info("In processing Bulk VM Event status %s", msg)
        time.sleep(3)
        if (not self.uplink_det_compl or
                phy_uplink not in self.ovs_vdp_obj_dict):
            # This condition shouldn't be hit as only when uplink is obtained,
            # save_uplink is called and that in turns calls this process_bulk.
            LOG.error("Uplink Port Event not received,"
                      "yet in bulk process")
            return
        ovs_vdp_obj = self.ovs_vdp_obj_dict[phy_uplink]
        for vm_dict in msg.msg_dict.get('vm_bulk_list'):
            if vm_dict['status'] == 'down':
                ovs_vdp_obj.pop_local_cache(vm_dict['port_uuid'],
                                            vm_dict['vm_mac'],
                                            vm_dict['net_uuid'],
                                            vm_dict['local_vlan'],
                                            vm_dict['vdp_vlan'],
                                            vm_dict['segmentation_id'])
            vm_msg = VdpQueMsg(constants.VM_MSG_TYPE,
                               port_uuid=vm_dict['port_uuid'],
                               vm_mac=vm_dict['vm_mac'],
                               net_uuid=vm_dict['net_uuid'],
                               segmentation_id=vm_dict['segmentation_id'],
                               status=vm_dict['status'],
                               oui=vm_dict['oui'],
                               phy_uplink=phy_uplink)
            self.process_vm_event(vm_msg, phy_uplink)

    def process_uplink_event(self, msg, phy_uplink):
        LOG.info("Received New uplink Msg %(msg)s for uplink %(uplink)s",
                 {'msg': msg.get_status(), 'uplink': phy_uplink})
        if msg.get_status() == 'up':
            ovs_exc_raised = False
            ovs_exc_reason = ""
            try:
                self.ovs_vdp_obj_dict[phy_uplink] = ovs_vdp.OVSNeutronVdp(
                    phy_uplink, msg.get_integ_br(), msg.get_ext_br(),
                    msg.get_root_helper(), self.vdp_vlan_change_cb)
            except Exception as exc:
                ovs_exc_reason = str(exc)
                LOG.error("OVS VDP Object creation failed %s",
                          str(ovs_exc_reason))
                ovs_exc_raised = True
            if (ovs_exc_raised or not self.ovs_vdp_obj_dict[phy_uplink].
                    is_lldpad_setup_done()):
                # Is there a way to delete the object??
                if not ovs_exc_reason:
                    uplink_fail_reason = (self.ovs_vdp_obj_dict[phy_uplink].
                                          get_uplink_fail_reason())
                else:
                    uplink_fail_reason = ovs_exc_reason
                LOG.error("UP Event Processing NOT Complete")
                self.err_que.enqueue(constants.Q_UPL_PRIO, msg)
                self.save_uplink(uplink=self.phy_uplink,
                                 fail_reason=uplink_fail_reason)
            else:
                self.uplink_det_compl = True
                veth_intf = (self.ovs_vdp_obj_dict[self.phy_uplink].
                             get_lldp_local_bridge_port())
                LOG.info("UP Event Processing Complete Saving uplink "
                         "%(ul)s and veth %(veth)s",
                         {'ul': self.phy_uplink, 'veth': veth_intf})
                self.save_uplink(uplink=self.phy_uplink, veth_intf=veth_intf)
                self.topo_disc.uncfg_intf(self.phy_uplink)
                self.topo_disc.cfg_intf(veth_intf,
                                        phy_interface=self.phy_uplink)
        elif msg.get_status() == 'down':
            # Free the object fixme(padkrish)
            if phy_uplink in self.ovs_vdp_obj_dict:
                self.ovs_vdp_obj_dict[phy_uplink].clear_obj_params()
            else:
                ovs_vdp.delete_uplink_and_flows(self.root_helper, self.br_ex,
                                                phy_uplink)
            self.save_uplink()
            self.topo_disc.uncfg_intf(self.veth_intf)
            self.topo_disc.cfg_intf(phy_uplink)

    def process_queue(self):
        LOG.info("Entered process_q")
        while True:
            prio, msg = self.que.dequeue()
            msg_type = msg.msg_type
            phy_uplink = msg.get_uplink()
            LOG.info("Msg dequeued type is %d", msg_type)
            try:
                if msg_type == constants.VM_MSG_TYPE:
                    self.process_vm_event(msg, phy_uplink)
                elif msg_type == constants.VM_BULK_SYNC_MSG_TYPE:
                    self.process_bulk_vm_event(msg, phy_uplink)
                elif msg_type == constants.UPLINK_MSG_TYPE:
                    try:
                        self.process_uplink_event(msg, phy_uplink)
                    except Exception as eu:
                        LOG.exception("Exception caught in process_uplink"
                                      " %s", str(eu))
                    self.process_uplink_ongoing = False
            except Exception as e:
                LOG.exception("Exception caught in process_q %s", str(e))

    def process_err_queue(self):
        LOG.info("Entered Err process_q")
        try:
            while self.err_que.is_not_empty():
                prio, msg = self.err_que.dequeue_nonblock()
                msg_type = msg.msg_type
                LOG.info("Msg dequeued from err queue type is %d",
                         msg_type)
                if msg_type == constants.UPLINK_MSG_TYPE:
                    self.que.enqueue(constants.Q_UPL_PRIO, msg)
        except Exception as e:
            LOG.exception("Exception caught in proc_err_que %s ", str(e))

    def start(self):
        # Spawn the thread
        # Pass the Que as last argument so that in case of exception, the
        # daemon can exit gracefully. fixme(padkrish)
        thr_q = utils.EventProcessingThread("VDP_Mgr", self, 'process_queue')
        thr_q.start()
        task_err_proc = utils.PeriodicTask(constants.ERR_PROC_INTERVAL,
                                           self.process_err_queue)
        task_err_proc.run()
        task_uplink = utils.PeriodicTask(constants.UPLINK_DET_INTERVAL,
                                         self.vdp_uplink_proc_top)
        task_uplink.run()

    def is_openstack_running(self):
        """Currently it just checks for the presence of both the bridges. """
        try:
            if (ovs_vdp.is_bridge_present(self.br_ex, self.root_helper) and
                    ovs_vdp.is_bridge_present(self.br_integ,
                                              self.root_helper)):
                return True
            else:
                return False
        except Exception as e:
            LOG.error("Exception in is_openstack_running %s", str(e))
            return False

    def vdp_uplink_proc_top(self):
        try:
            self.vdp_uplink_proc()
        except Exception as e:
            LOG.error("VDP uplink proc exception %s", e)

    def save_uplink(self, uplink="", veth_intf="", fail_reason=""):
        context = {}
        # If uplink physical interface is a part of bond, then this function
        # will be called with uplink=bond0, as an example
        memb_port_list = sys_utils.get_member_ports(uplink)
        args = jsonutils.dumps({'agent': self.host_id, 'uplink': uplink,
                                'veth_intf': veth_intf,
                                'memb_port_list': memb_port_list,
                                'fail_reason': fail_reason})
        msg = self.rpc_clnt.make_msg('save_uplink', context, msg=args)
        try:
            resp = self.rpc_clnt.call(msg)
            return resp
        except rpc.MessagingTimeout:
            LOG.error("RPC timeout: Failed to save link name on the "
                      "server")

    def _fill_topology_cfg(self, topo_dict):
        """Fills the extra configurations in the topology. """
        cfg_dict = {}
        if topo_dict.bond_member_ports is not None:
            cfg_dict.update({'bond_member_ports':
                             topo_dict.bond_member_ports})
        if topo_dict.bond_interface is not None:
            cfg_dict.update({'bond_interface':
                             topo_dict.bond_interface})
        return cfg_dict

    def save_topo_disc_params(self, intf, topo_disc_obj):
        context = {}
        topo_cfg = self._fill_topology_cfg(topo_disc_obj)
        args = jsonutils.dumps(
            {'host': self.host_id, 'protocol_interface': intf,
             'heartbeat': time.ctime(),
             'phy_interface': topo_disc_obj.phy_interface,
             'remote_evb_cfgd': topo_disc_obj.remote_evb_cfgd,
             'remote_evb_mode': topo_disc_obj.remote_evb_mode,
             'remote_mgmt_addr': topo_disc_obj.remote_mgmt_addr,
             'remote_system_desc': topo_disc_obj.remote_system_desc,
             'remote_system_name': topo_disc_obj.remote_system_name,
             'remote_port': topo_disc_obj.remote_port,
             'remote_chassis_id_mac': topo_disc_obj.remote_chassis_id_mac,
             'remote_port_id_mac': topo_disc_obj.remote_port_id_mac,
             'configurations': jsonutils.dumps(topo_cfg)})
        msg = self.rpc_clnt.make_msg('save_topo_disc_params', context,
                                     msg=args)
        try:
            resp = self.rpc_clnt.call(msg)
            return resp
        except rpc.MessagingTimeout:
            LOG.error("RPC timeout: Failed to send topo disc on the server")

    def uplink_bond_intf_process(self):
        """Process the case when uplink interface becomes part of a bond.

        This is called to check if the phy interface became a part of the
        bond. If the below condition is True, this means, a physical
        interface that was not a part of a bond was earlier discovered as
        uplink and now that interface became part of the bond.
        Usually, this doesn't happen as LLDP and in turn this function will
        first detect a 'down' followed by an 'up'. When regular interface
        becomes part of bond, it's rare for it to hit this 'normal' case.
        But, still providing the functionality if it happens.
        The following is done :
        a. Bring down the physical interface by sending a 'down' event
        b. Add the bond interface by sending an 'up' event
        Consquently, when bond is added that will be assigned to
        self.phy_uplink. Then, the below condition will be False. i.e..
        'get_bond_intf' will return False, when the argument is 'bond0'.
        """
        bond_intf = sys_utils.get_bond_intf(self.phy_uplink)
        if bond_intf is None:
            return False
        self.save_uplink(
            fail_reason=constants.port_transition_bond_down_reason)
        self.process_uplink_ongoing = True
        upl_msg = VdpQueMsg(constants.UPLINK_MSG_TYPE, status='down',
                            phy_uplink=self.phy_uplink,
                            br_int=self.br_integ, br_ex=self.br_ex,
                            root_helper=self.root_helper)
        self.que.enqueue(constants.Q_UPL_PRIO, upl_msg)
        self.phy_uplink = None
        self.veth_intf = None
        self.uplink_det_compl = False

        # No veth interface
        self.save_uplink(
            uplink=bond_intf,
            fail_reason=constants.port_transition_bond_up_reason)
        self.phy_uplink = bond_intf
        self.process_uplink_ongoing = True
        upl_msg = VdpQueMsg(constants.UPLINK_MSG_TYPE, status='up',
                            phy_uplink=self.phy_uplink,
                            br_int=self.br_integ, br_ex=self.br_ex,
                            root_helper=self.root_helper)
        self.que.enqueue(constants.Q_UPL_PRIO, upl_msg)
        return True

    def check_periodic_bulk_vm_notif_rcvd(self):
        """Bulk VM check handler called from periodic uplink detection.

        This gets called by the 'normal' stage of uplink detection.
        The bulk VM event sends all the VM's running in this agent.
        Sometimes during upgrades, it was found that due to some race
        condition, the server does not send the Bulk VM event.
        Whenever, a save_uplink is done by the agent, the server sends
        the Bulk VM event.
        If Bulk VM event is not received after few attempts, save_uplink is
        done to request the Bulk VM list.
        It's not protected with a mutex, since worst case,
        Bulk VM event will be sent twice, which is not that bad. When
        uplink is detected for the first time, it will hit the below
        else case and there a save_uplink is anyways done.
        """
        if not self.bulk_vm_rcvd_flag:
            if self.bulk_vm_check_cnt >= 1:
                self.bulk_vm_check_cnt = 0
                self.save_uplink(uplink=self.phy_uplink,
                                 veth_intf=self.veth_intf)
                LOG.info("Doing save_uplink again to request "
                         "Bulk VM's")
            else:
                LOG.info("Bulk VM not received, incrementing count")
                self.bulk_vm_check_cnt += 1

    def static_uplink_detect(self, veth):
        """Return the static uplink based on argument passed.

        The very first time, this function is called, it returns the uplink
        port read from a file.
        After restart, when this function is called the first time, it
        returns 'normal' assuming a veth is passed to this function which will
        be the case if uplink processing is successfully done.
        If user modified the uplink configuration and restarted, a 'down'
        will be returned to clear the old uplink.
        """
        LOG.info("In static_uplink_detect %(veth)s", {'veth': veth})
        if self.static_uplink_first:
            self.static_uplink_first = False
            if self.phy_uplink is not None and (
               self.phy_uplink != self.static_uplink_port):
                return 'down'
        if veth is None:
            return self.static_uplink_port
        else:
            return 'normal'

    def vdp_uplink_proc(self):
        """Periodic handler to detect the uplink interface to the switch.

        -> restart_uplink_called: should be called by agent initially to set
        the stored uplink and veth from DB
        -> process_uplink_ongoing: Will be set when uplink message is enqueue
        and reset when dequeued and processed completely
        -> uplink_det_compl: Will be set to True when a valid uplink is
        detected and object created. Will be reset when uplink is down
        -> phy_uplink: Is the uplink interface
        -> veth_intf : Signifies the veth interface.
        """
        LOG.info("In Periodic Uplink Task")
        if not self.is_os_run:
            if not self.is_openstack_running():
                LOG.info("OpenStack is not running")
                return
            else:
                self.is_os_run = True
        if not self.restart_uplink_called or self.process_uplink_ongoing:
            LOG.info("Uplink before restart not refreshed yet..states "
                     "%(ruc)d %(puo)d",
                     {'ruc': self.restart_uplink_called,
                      'puo': self.process_uplink_ongoing})
            return
        if self.phy_uplink is not None:
            if (self.uplink_det_compl and (
               self.phy_uplink not in self.ovs_vdp_obj_dict)):
                LOG.error("Not Initialized for phy %s", self.phy_uplink)
                return
            if self.phy_uplink in self.ovs_vdp_obj_dict:
                self.veth_intf = (self.ovs_vdp_obj_dict[self.phy_uplink].
                                  get_lldp_local_bridge_port())
            # The below logic has a bug when agent is started
            # and openstack is not running fixme(padkrish)
            else:
                if self.veth_intf is None:
                    LOG.error("Incorrect state, Bug")
                    return
        if self.static_uplink:
            ret = self.static_uplink_detect(self.veth_intf)
        else:
            ret = uplink_det.detect_uplink(self.veth_intf)
        if ret is 'down':
            if self.phy_uplink is None:
                LOG.error("Wrong status down")
                return
            # Call API to set the uplink as "" DOWN event
            self.uplink_down_cnt = self.uplink_down_cnt + 1
            if not self.static_uplink and (
               self.uplink_down_cnt < constants.UPLINK_DOWN_THRES):
                return
            self.process_uplink_ongoing = True
            upl_msg = VdpQueMsg(constants.UPLINK_MSG_TYPE,
                                status='down',
                                phy_uplink=self.phy_uplink,
                                br_int=self.br_integ, br_ex=self.br_ex,
                                root_helper=self.root_helper)
            self.que.enqueue(constants.Q_UPL_PRIO, upl_msg)
            self.phy_uplink = None
            self.veth_intf = None
            self.uplink_det_compl = False
            self.uplink_down_cnt = 0
        elif ret is None:
            if self.veth_intf is not None:
                LOG.error("Wrong status None")
                return
            # Call API to set the uplink as "" Uplink not discovered yet
            self.save_uplink(fail_reason=constants.uplink_undiscovered_reason)
        elif ret is 'normal':
            if self.veth_intf is None:
                LOG.error("Wrong status Normal")
                return
            # Uplink already discovered, nothing to be done here
            # Resetting it back, happens when uplink was down for a very short
            # time and no need to remove flows
            self.uplink_down_cnt = 0
            bond_det = self.uplink_bond_intf_process()
            # Revisit this logic.
            # If uplink detection fails, it will be put in Error queue, which
            # will dequeue and put it back in the main queue
            # At the same time this periodic task will also hit this normal
            # state and will put the message in main queue. fixme(padkrish)
            # The below lines are put here because after restart when
            # eth/veth are passed to uplink script, it will return normal
            # But OVS object would not have been created for the first time,
            # so the below lines ensures it's done.
            if not self.uplink_det_compl and not bond_det:
                if self.phy_uplink is None:
                    LOG.error("Incorrect state, bug")
                    return
                self.process_uplink_ongoing = True
                upl_msg = VdpQueMsg(constants.UPLINK_MSG_TYPE,
                                    status='up',
                                    phy_uplink=self.phy_uplink,
                                    br_int=self.br_integ, br_ex=self.br_ex,
                                    root_helper=self.root_helper)
                self.que.enqueue(constants.Q_UPL_PRIO, upl_msg)
                # yield
                LOG.info("Enqueued Uplink Msg from normal")
            self.check_periodic_bulk_vm_notif_rcvd()
        else:
            LOG.info("In Periodic Uplink Task uplink found %s", ret)
            bond_intf = sys_utils.get_bond_intf(ret)
            if bond_intf is not None:
                ret = bond_intf
                LOG.info("Interface %(memb)s part of bond %(bond)s" %
                         {'memb': ret, 'bond': bond_intf})
            # Call API to set the uplink as ret
            self.save_uplink(uplink=ret, veth_intf=self.veth_intf)
            self.phy_uplink = ret
            self.process_uplink_ongoing = True
            upl_msg = VdpQueMsg(constants.UPLINK_MSG_TYPE,
                                status='up',
                                phy_uplink=self.phy_uplink,
                                br_int=self.br_integ, br_ex=self.br_ex,
                                root_helper=self.root_helper)
            self.que.enqueue(constants.Q_UPL_PRIO, upl_msg)
            # yield
            LOG.info("Enqueued Uplink Msg")

    def vdp_vm_event(self, vm_dict_list):
        if isinstance(vm_dict_list, list):
            vm_msg = VdpQueMsg(constants.VM_BULK_SYNC_MSG_TYPE,
                               vm_bulk_list=vm_dict_list,
                               phy_uplink=self.phy_uplink)
            self.bulk_vm_rcvd_flag = True
        else:
            vm_dict = vm_dict_list
            LOG.info("Obtained VM event Enqueueing Status %(status)s "
                     "MAC %(mac)s uuid %(uuid)s oui %(oui)s",
                     {'status': vm_dict['status'], 'mac': vm_dict['vm_mac'],
                      'uuid': vm_dict['net_uuid'], 'oui': vm_dict['oui']})
            vm_msg = VdpQueMsg(constants.VM_MSG_TYPE,
                               port_uuid=vm_dict['port_uuid'],
                               vm_mac=vm_dict['vm_mac'],
                               net_uuid=vm_dict['net_uuid'],
                               segmentation_id=vm_dict['segmentation_id'],
                               status=vm_dict['status'],
                               oui=vm_dict['oui'],
                               phy_uplink=self.phy_uplink)
        self.que.enqueue(constants.Q_VM_PRIO, vm_msg)

    def is_uplink_received(self):
        """Returns whether uplink information is received after restart.

        Not protecting this with a mutex, since this gets called inside the
        loop from dfa_agent and having a mutex is a overkill. Worst case,
        during multiple restarts on server and when the corner case is hit,
        this may return an incorrect value of False when _dfa_uplink_restart
        is at the middle of execution. Returning an incorrect value of False,
        may trigger an RPC to the server to retrieve the uplink one extra time.
        _dfa_uplink_restart will not get executed twice, since that is anyway
        protected with a mutex.
        """
        return self.restart_uplink_called

    def dfa_uplink_restart(self, uplink_dict):
        try:
            with self.vdp_mgr_lock:
                if not self.restart_uplink_called:
                    self._dfa_uplink_restart(uplink_dict)
        except Exception as exc:
            LOG.error("Exception in dfa_uplink_restart %s" % str(exc))

    def _dfa_uplink_restart(self, uplink_dict):
        LOG.info("Obtained uplink after restart %s ", uplink_dict)
        # This shouldn't happen
        if self.phy_uplink is not None:
            LOG.error("Uplink detection already done %s", self.phy_uplink)
            return
        uplink = uplink_dict.get('uplink')
        veth_intf = uplink_dict.get('veth_intf')
        # Logic is as follows:
        # If DB didn't have any uplink it means it's not yet detected or down
        # if DB has uplink and veth, then no need to scan all ports we can
        # start with this veth.
        # If uplink has been removed or modified during restart, then a
        # down will be returned by uplink detection code and it will be
        # removed then.
        # If DB has uplink, but no veth, it's an error condition and in
        # which case remove the uplink port from bridge and start fresh
        if uplink is None or len(uplink) == 0:
            LOG.info("uplink not discovered yet")
            self.restart_uplink_called = True
            return
        if veth_intf is not None and len(veth_intf) != 0:
            LOG.info("veth interface is already added, %(ul)s %(veth)s",
                     {'ul': uplink, 'veth': veth_intf})
            self.phy_uplink = uplink
            self.veth_intf = veth_intf
            self.restart_uplink_called = True
            return
        LOG.info("Error case removing the uplink %s from bridge", uplink)
        ovs_vdp.delete_uplink_and_flows(self.root_helper, self.br_ex, uplink)
        self.restart_uplink_called = True



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from neutron.api.v2 import attributes as attr
from neutron.api.v2 import resource_helper
from neutron_lib.api import converters as conv
from neutron_lib import exceptions

from networking_cisco._i18n import _
from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.common import utils


NEUTRON_VERSION = bc.NEUTRON_VERSION
NEUTRON_NEWTON_VERSION = bc.NEUTRON_NEWTON_VERSION

ROUTERTYPE = 'routertype'
ROUTERTYPE_ALIAS = ROUTERTYPE
TYPE_ATTR = ROUTERTYPE + ':id'
ROUTER_TYPES = ROUTERTYPE + 's'

RESOURCE_ATTRIBUTE_MAP = {
    ROUTER_TYPES: {
        'id': {'allow_post': True, 'allow_put': False,
               'validate': {'type:uuid_or_none': None}, 'is_visible': True,
               'default': None, 'primary_key': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:string': None}, 'is_visible': True,
                 'default': ''},
        'description': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:string_or_none': None},
                        'is_visible': True, 'default': None},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'required_by_policy': True, 'is_visible': True},
        'template_id': {'allow_post': True, 'allow_put': False,
                        'required_by_policy': True,
                        'validate': {'type:uuid': None}, 'is_visible': True},
        'ha_enabled_by_default': {'allow_post': True, 'allow_put': True,
                                  'convert_to': conv.convert_to_boolean,
                                  'validate': {'type:boolean': None},
                                  'default': False, 'is_visible': True},
        'shared': {'allow_post': True, 'allow_put': False,
                   'convert_to': conv.convert_to_boolean,
                   'validate': {'type:boolean': None}, 'default': True,
                   'is_visible': True},
        #TODO(bobmel): add HA attribute: One of None, 'GPLB', 'VRRP', or 'HSRP'
        'slot_need': {'allow_post': True, 'allow_put': True,
                      'validate': {'type:non_negative': None},
                      'convert_to': conv.convert_to_int,
                      'default': 0, 'is_visible': True},
        'scheduler': {'allow_post': True, 'allow_put': False,
                      'required_by_policy': True,
                      'convert_to': utils.convert_validate_driver_class,
                      'is_visible': True},
        'driver': {'allow_post': True, 'allow_put': False,
                   'required_by_policy': True,
                   'convert_to': utils.convert_validate_driver_class,
                   'is_visible': True},
        'cfg_agent_service_helper': {
            'allow_post': True, 'allow_put': False,
            'required_by_policy': True,
            'convert_to': utils.convert_validate_driver_class,
            'is_visible': True},
        'cfg_agent_driver': {'allow_post': True, 'allow_put': False,
                             'required_by_policy': True,
                             'convert_to': utils.convert_validate_driver_class,
                             'is_visible': True},
    }
}

EXTENDED_ATTRIBUTES_2_0 = {
    'routers': {
        TYPE_ATTR: {'allow_post': True, 'allow_put': True,
                    'validate': {'type:string': None},
                    'default': bc.constants.ATTR_NOT_SPECIFIED,
                    'is_visible': True},
    }
}


class Routertype(bc.extensions.ExtensionDescriptor):
    """Extension class to define different types of Neutron routers.

    This class is used by Neutron's extension framework to support
    definition of different types of Neutron Routers.

    Attribute 'router_type:id' is the uuid or name of a certain router type.
    It can be set during creation of Neutron router. If a Neutron router is
    moved (by admin user) to a hosting device of a different hosting device
    type, the router type of the Neutron router will also change. Non-admin
    users can request that a Neutron router's type is changed.

    To create a router of router type <name>:

       (shell) router-create <router_name> --router_type:id <uuid_or_name>
    """

    @classmethod
    def get_name(cls):
        return "Router types for routing service"

    @classmethod
    def get_alias(cls):
        return ROUTERTYPE_ALIAS

    @classmethod
    def get_description(cls):
        return "Introduces router types for Neutron Routers"

    @classmethod
    def get_namespace(cls):
        return "http://docs.openstack.org/ext/" + ROUTERTYPE + "/api/v2.0"

    @classmethod
    def get_updated(cls):
        return "2014-02-07T10:00:00-00:00"

    @classmethod
    def get_resources(cls):
        """Returns Ext Resources."""
        plural_mappings = resource_helper.build_plural_mappings(
            {}, RESOURCE_ATTRIBUTE_MAP)
        if NEUTRON_VERSION.version[0] <= NEUTRON_NEWTON_VERSION.version[0]:
            attr.PLURALS.update(plural_mappings)
        return resource_helper.build_resource_info(plural_mappings,
                                                   RESOURCE_ATTRIBUTE_MAP,
                                                   bc.constants.L3)

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}


# router_type exceptions
class RouterTypeInUse(exceptions.InUse):
    message = _("Router type %(id)s in use.")


class RouterTypeNotFound(exceptions.NotFound):
    message = _("Router type %(id)s does not exist")


class MultipleRouterTypes(exceptions.NeutronException):
    message = _("Multiple router type with same name %(name)s exist. Id "
                "must be used to specify router type.")


class SchedulerNotFound(exceptions.NetworkNotFound):
    message = _("Scheduler %(scheduler)s does not exist")


class RouterTypeAlreadyDefined(exceptions.NeutronException):
    message = _("Router type %(type)s already exists")


class NoSuchHostingDeviceTemplateForRouterType(exceptions.NeutronException):
    message = _("No hosting device template with id %(type)s exists")


class HostingDeviceTemplateUsedByRouterType(exceptions.NeutronException):
    message = _("Router type %(type)s already defined for Hosting device "
                "template with id %(type)s")


class RouterTypeHasRouters(exceptions.NeutronException):
    message = _("Router type %(type)s cannot be deleted since routers "
                "of that type exists")


class RoutertypePluginBase(object):
    """REST API to manage router types.

    All methods except listing require admin context.
    """
    @abc.abstractmethod
    def create_routertype(self, context, routertype):
        """Creates a router type.
         Also binds it to the specified hosting device template.
         """
        pass

    @abc.abstractmethod
    def update_routertype(self, context, id, routertype):
        """Updates a router type."""
        pass

    @abc.abstractmethod
    def delete_routertype(self, context, id):
        """Deletes a router type."""
        pass

    @abc.abstractmethod
    def get_routertype(self, context, id, fields=None):
        """Lists defined router type."""
        pass

    @abc.abstractmethod
    def get_routertypes(self, context, filters=None, fields=None,
                        sorts=None, limit=None, marker=None,
                        page_reverse=False):
        """Lists defined router types."""
        pass



# Copyright 2015 Cisco Systems, Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_cisco import backwards_compatibility as bc
from networking_cisco.ml2_drivers.nexus import nexus_models_v2  # noqa
from networking_cisco.ml2_drivers.ucsm import ucsm_model  # noqa
from networking_cisco.plugins.cisco.db.device_manager import hd_models  # noqa
from networking_cisco.plugins.cisco.db.l3 import ha_db  # noqa
from networking_cisco.plugins.cisco.db.l3 import l3_models  # noqa


def get_metadata():
    return bc.model_base.BASEV2.metadata



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import mock

from oslo_config import cfg
import oslo_messaging
from oslo_utils import uuidutils

from neutron.tests import base

from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.cfg_agent import cfg_agent
from networking_cisco.plugins.cisco.cfg_agent import cfg_exceptions
from networking_cisco.plugins.cisco.cfg_agent.device_drivers import driver_mgr
from networking_cisco.plugins.cisco.cfg_agent.service_helpers import (
    routing_svc_helper)
from networking_cisco.plugins.cisco.common import (cisco_constants as
                                                   c_constants)
from networking_cisco.plugins.cisco.extensions import ha
from networking_cisco.plugins.cisco.extensions import routerrole
from networking_cisco.tests.unit.cisco.cfg_agent import cfg_agent_test_support


_uuid = uuidutils.generate_uuid
HOST = 'myhost'
FAKE_ID = _uuid()
HA_INFO = 'ha_info'


class TestRouterInfo(base.BaseTestCase,
                     cfg_agent_test_support.CfgAgentTestSupportMixin):

    def setUp(self):
        super(TestRouterInfo, self).setUp()
        self.ex_gw_port = {'id': _uuid(),
                           'network_id': _uuid(),
                           'fixed_ips': [{'ip_address': '19.4.4.4',
                                          'subnet_id': _uuid()}],
                           'subnets': [{'cidr': '19.4.4.0/24',
                                        'gateway_ip': '19.4.4.1'}]}
        self.router = {'id': _uuid(),
                       'enable_snat': True,
                       'routes': [],
                       'gw_port': self.ex_gw_port}

    def test_router_info_create(self):
        router_id = _uuid()
        fake_router = {}
        ri = routing_svc_helper.RouterInfo(router_id, fake_router)

        self.assertTrue(ri.router_name().endswith(router_id))

    def test_router_info_create_with_router(self):
        router_id = _uuid()
        ri = routing_svc_helper.RouterInfo(router_id, self.router)
        self.assertTrue(ri.router_name().endswith(router_id))
        self.assertEqual(ri.router, self.router)
        self.assertEqual(ri._router, self.router)
        self.assertTrue(ri.snat_enabled)
        self.assertIsNone(ri.ex_gw_port)

    def test_router_info_create_snat_disabled(self):
        router_id = _uuid()
        self.router['enable_snat'] = False
        ri = routing_svc_helper.RouterInfo(router_id, self.router)
        self.assertFalse(ri.snat_enabled)


class TestBasicRoutingOperations(
        base.BaseTestCase, cfg_agent_test_support.CfgAgentTestSupportMixin):

    def setUp(self):
        super(TestBasicRoutingOperations, self).setUp()
        self.conf = cfg.ConfigOpts()
        self.conf.register_opts(bc.core_opts)
        self.conf.register_opts(cfg_agent.OPTS, "cfg_agent")
        sn_id = _uuid()
        self.ex_gw_port = {'id': _uuid(),
                           'network_id': _uuid(),
                           'admin_state_up': True,
                           'fixed_ips': [{'ip_address': '19.4.4.4',
                                         'subnet_id': sn_id}],
                           'subnets': [{'id': sn_id, 'cidr': '19.4.4.0/24',
                                        'gateway_ip': '19.4.4.1'}]}
        self.hosting_device = {'id': "100",
                               'name': "VM_template",
                               'booting_time': 300,
                               'host_category': "VM",
                               'management_ip_address': '20.0.0.5',
                               'protocol_port': 22,
                               'credentials': {'credentials_id': 'creds_uuid1',
                                               'user_name': 'user',
                                               'password': '********'},
                               }
        self.router = {
            'id': _uuid(),
            'name': 'router1',
            'enable_snat': True,
            'admin_state_up': True,
            'routes': [],
            'gw_port': self.ex_gw_port,
            routerrole.ROUTER_ROLE_ATTR: None,
            'hosting_device': self.hosting_device,
            'router_type': {
                'cfg_agent_driver': 'networking_cisco.plugins.cisco.'
                                    'cfg_agent.device_drivers.iosxe.'
                                    'iosxe_routing_driver.IosXeRoutingDriver'}}

        self.agent = mock.Mock()

        #Patches & Mocks

        self.l3pluginApi_cls_p = mock.patch(
            'networking_cisco.plugins.cisco.cfg_agent.service_helpers.'
            'routing_svc_helper.CiscoRoutingPluginApi')
        l3plugin_api_cls = self.l3pluginApi_cls_p.start()
        self.plugin_api = mock.Mock()
        l3plugin_api_cls.return_value = self.plugin_api
        self.plugin_api.get_routers = mock.MagicMock()
        self.plugin_api.get_router_ids = mock.MagicMock()
        self.looping_call_p = mock.patch(
            'oslo_service.loopingcall.FixedIntervalLoopingCall')
        self.looping_call_p.start()
        mock.patch('neutron.common.rpc.create_connection').start()

        self.routing_helper = routing_svc_helper.RoutingServiceHelper(
            HOST, self.conf, self.agent)
        self.routing_helper._internal_network_added = mock.Mock()
        self.routing_helper._external_gateway_added = mock.Mock()
        self.routing_helper._internal_network_removed = mock.Mock()
        self.routing_helper._external_gateway_removed = mock.Mock()
        self.routing_helper._enable_router_interface = mock.Mock()
        self.routing_helper._disable_router_interface = mock.Mock()
        self.driver = self._mock_driver_and_hosting_device(
            self.routing_helper)

    def _mock_driver_and_hosting_device(self, svc_helper):
        svc_helper._dev_status.is_hosting_device_reachable = mock.MagicMock(
            return_value=True)
        driver = mock.MagicMock()
        svc_helper._drivermgr.get_driver = mock.Mock(return_value=driver)
        svc_helper._drivermgr.set_driver = mock.Mock(return_value=driver)
        return driver

    def _reset_mocks(self):
        self.routing_helper._process_router_floating_ips.reset_mock()
        self.routing_helper._internal_network_added.reset_mock()
        self.routing_helper._external_gateway_added.reset_mock()
        self.routing_helper._internal_network_removed.reset_mock()
        self.routing_helper._external_gateway_removed.reset_mock()
        self.routing_helper._enable_router_interface.reset_mock()
        self.routing_helper._disable_router_interface.reset_mock()

    def test_process_router_throw_config_error(self):
        snip_name = 'CREATE_SUBINTERFACE'
        e_type = 'Fake error'
        e_tag = 'Fake error tag'
        confstr = 'Fake conf str'
        params = {'snippet': snip_name, 'type': e_type, 'tag': e_tag,
                  'confstr': confstr, 'dev_id': 'FAKE_ID', 'ip': 'FAKE_IP'}
        self.routing_helper._internal_network_added.side_effect = (
            cfg_exceptions.IOSXEConfigException(**params))
        router, ports = self.prepare_router_data()
        ri = routing_svc_helper.RouterInfo(router['id'], router)
        self.assertRaises(cfg_exceptions.IOSXEConfigException,
                          self.routing_helper._process_router, ri)

    def test_process_router_throw_session_close(self):
        class SessionCloseError(Exception):
            pass

        self.routing_helper._internal_network_added.side_effect = (
            SessionCloseError("Simulate SessionCloseError"))
        router, ports = self.prepare_router_data()
        ri = routing_svc_helper.RouterInfo(router['id'], router)
        self.assertRaises(SessionCloseError,
                          self.routing_helper._process_router, ri)

    def test_process_router_throw_no_ip_address_on_subnet_error(self):
        router, ports = self.prepare_router_data()
        # change subnet_id to something that does not exist
        router['gw_port']['subnets'][0]['id'] = 'fake_uuid'
        ri = routing_svc_helper.RouterInfo(router['id'], router)
        self.assertRaises(routing_svc_helper.IPAddressMissingException,
                          self.routing_helper._process_router, ri)

    def _test_process_router_throw_multiple_ipv4_subnets_error(self):
        router, ports = self.prepare_router_data(num_ext_subnets=2,
                                                 is_global=True)
        # make router a regular one which triggers the error condition
        router[routerrole.ROUTER_ROLE_ATTR] = None
        ri = routing_svc_helper.RouterInfo(router['id'], router)
        return ri, router

    def test_process_router_throw_multiple_ipv4_subnets_error(self):
        ri, router = (
            self._test_process_router_throw_multiple_ipv4_subnets_error())
        self.assertRaises(routing_svc_helper.MultipleIPv4SubnetsException,
                          self.routing_helper._process_router, ri)

    def _test_router_admin_port_state(self, router, ri, ex_gw_port):
        # change the router admin_state_up to false
        router['admin_state_up'] = False
        ri.router = router
        self.routing_helper._process_router(ri)
        self.routing_helper._disable_router_interface.assert_called_with(
            ri)
        self.assertFalse(self.routing_helper._enable_router_interface.called)
        self._reset_mocks()

        # Change router admin_state_up to True, and set the router port
        # admin_state_up to false
        router['admin_state_up'] = True
        router['gw_port']['admin_state_up'] = False
        ri.router = router
        self.routing_helper._process_router(ri)
        self.routing_helper._disable_router_interface.assert_called_with(
            ri, ex_gw_port)
        self._reset_mocks()

        # Change admin_state_up for Router and router port to True
        router['gw_port']['admin_state_up'] = True
        ri.router = router
        self.routing_helper._process_router(ri)
        self.routing_helper._enable_router_interface.assert_any_call(
            ri, ex_gw_port)
        self.routing_helper._enable_router_interface.assert_any_call(
            ri, router[bc.constants.INTERFACE_KEY][0])
        self.assertFalse(self.routing_helper._disable_router_interface.called)
        self._reset_mocks()

    def _verify_operations(self, added, removed):

        def _verify_ports(expected_pids, actual_ports):
            num_elem = len(expected_pids)
            self.assertEqual(num_elem, len(actual_ports))
            for i in range(num_elem):
                self.assertEqual(expected_pids[i], actual_ports[i]['id'])

        actual_added = self.routing_helper._internal_network_added
        actual_removed = self.routing_helper._internal_network_removed
        self.assertEqual(len(added), actual_added.call_count)
        i = 0
        for expected in added:
            port_dict = actual_added.call_args_list[i][0][1]
            actual_details = port_dict['change_details']
            self.assertEqual(expected['p_id'], port_dict['id'])
            _verify_ports(expected['new_port_pids'],
                          actual_details['new_ports'])
            _verify_ports(expected['current_port_pids'],
                          actual_details['current_ports'])
            _verify_ports(expected['old_port_pids'],
                          actual_details['old_ports'])
            _verify_ports(expected['former_port_pids'],
                          actual_details['former_ports'])
            i += 1
        self.assertEqual(len(removed), actual_removed.call_count)
        i = 0
        for expected in removed:
            port_dict = actual_removed.call_args_list[i][0][1]
            actual_details = port_dict['change_details']
            self.assertEqual(expected['p_id'], port_dict['id'])
            _verify_ports(expected['new_port_pids'],
                          actual_details['new_ports'])
            _verify_ports(expected['current_port_pids'],
                          actual_details['current_ports'])
            _verify_ports(expected['old_port_pids'],
                          actual_details['old_ports'])
            _verify_ports(expected['former_port_pids'],
                          actual_details['former_ports'])
            i += 1
        self.routing_helper._internal_network_added.reset_mock()
        self.routing_helper._internal_network_removed.reset_mock()

    def test_process_router_multiple_ports_on_same_multiple_subnet_network(
            self):
        router, ports = self.prepare_router_data(
            set_gateway=True, num_internal_ports=1, same_internal_nw=True)
        port_1 = ports[0]
        p1_id = port_1['id']
        nw_uuid = ports[0]['network_id']
        vlan_tag = ports[0]['hosting_info']['segmentation_id']
        ri = routing_svc_helper.RouterInfo(router['id'], router=router)
        # process with router attached to one subnet
        self.routing_helper._process_router(ri)
        # ----> VERIFY HERE: [.35]
        # attach router to second subnet on same network, subnet should become
        self._verify_operations(
            [{'p_id': p1_id,
              'new_port_pids': [p1_id],
              'current_port_pids': [p1_id],
              'old_port_pids': [],
              'former_port_pids': []}], [])
        # primary since 32 < 35 determines sorting order
        port_2 = self.create_router_port(
            nw_uuid, vlan_tag, 32, 1, router['id'],
            mac_address='ac:22:de:ad:be:ef')
        ports.append(port_2)
        p2_id = port_2['id']
        # process again
        self.routing_helper._process_router(ri)
        # ----> VERIFY HERE: [.32], .35
        self._verify_operations(
            [{'p_id': p2_id,
              'new_port_pids': [p2_id],
              'current_port_pids': [p2_id, p1_id],
              'old_port_pids': [],
              'former_port_pids': [p1_id]}], [])
        # attach router to third subnet on same network, subnet should become
        # secondary since 36 > 32 determines sorting order
        port_3 = self.create_router_port(
            nw_uuid, vlan_tag, 36, 1, router['id'],
            mac_address='ac:33:de:ad:be:ef')
        ports.append(port_3)
        p3_id = port_3['id']
        # process again
        self.routing_helper._process_router(ri)
        # ----> VERIFY HERE: [.32], .35, .36
        self._verify_operations(
            [{'p_id': p3_id,
              'new_port_pids': [p3_id],
              'current_port_pids': [p2_id, p1_id, p3_id],
              'old_port_pids': [],
              'former_port_pids': [p2_id, p1_id]}], [])
        # attach router to fourth subnet on same network, subnet should become
        # secondary since 36 > 32 determines sorting order
        port_4 = self.create_router_port(
            nw_uuid, vlan_tag, 37, 1, router['id'],
            mac_address='ac:44:de:ad:be:ef')
        ports.append(port_4)
        p4_id = port_4['id']
        # process again
        self.routing_helper._process_router(ri)
        # ----> VERIFY HERE: [.32], .35, .36, .37
        self._verify_operations(
            [{'p_id': p4_id,
             'new_port_pids': [p4_id],
             'current_port_pids': [p2_id, p1_id, p3_id, p4_id],
             'old_port_pids': [],
             'former_port_pids': [p2_id, p1_id, p3_id]}], [])
        # detach router from first subnet on same network, primary should
        # remain the same since 35 > 32 determines sorting order
        del ports[0]
        # process again
        self.routing_helper._process_router(ri)
        # ----> VERIFY HERE: [.32], .36, .37
        self._verify_operations(
            [],
            [{'p_id': p1_id,
              'new_port_pids': [],
              'current_port_pids': [p2_id, p3_id, p4_id],
              'old_port_pids': [p1_id],
              'former_port_pids': [p2_id, p1_id, p3_id, p4_id]}])
        # detach router from second subnet on same network, primary must
        # change since 32 < 36 determines sorting order
        del ports[0]
        # process again
        self.routing_helper._process_router(ri)
        # ----> VERIFY HERE: [.36], .37
        self._verify_operations(
            [],
            [{'p_id': p2_id,
              'new_port_pids': [],
              'current_port_pids': [p3_id, p4_id],
              'old_port_pids': [p2_id],
              'former_port_pids': [p2_id, p3_id, p4_id]}])
        # attach router to fifth subnet on same network, subnet should become
        # primary since 33 < 36 determines sorting order
        port_5 = self.create_router_port(
            nw_uuid, vlan_tag, 33, 1, router['id'],
            mac_address='ac:55:de:ad:be:ef')
        ports.append(port_5)
        p5_id = port_5['id']
        # detach router from third subnet on same network, primary must
        # not change since 37 > 33
        del ports[0]
        # process again
        self.routing_helper._process_router(ri)
        # ----> VERIFY HERE: [.33], .37
        self._verify_operations(
            [{'p_id': p5_id,
              'new_port_pids': [p5_id],
              'current_port_pids': [p5_id, p4_id],
              'old_port_pids': [p3_id],
              'former_port_pids': [p3_id, p4_id]}],
            [{'p_id': p3_id,
              'new_port_pids': [p5_id],
              'current_port_pids': [p5_id, p4_id],
              'old_port_pids': [p3_id],
              'former_port_pids': [p3_id, p4_id]}])
        # attach router to sixth subnet on same network, subnet should become
        # primary since 31 < 33 determines sorting order
        port_6 = self.create_router_port(
            nw_uuid, vlan_tag, 31, 1, router['id'],
            mac_address='ac:66:de:ad:be:ef')
        ports.append(port_6)
        p6_id = port_6['id']
        self.routing_helper._process_router(ri)
        # ----> VERIFY HERE: [.31], .33, .37
        self._verify_operations(
            [{'p_id': p6_id,
              'new_port_pids': [p6_id],
              'current_port_pids': [p6_id, p5_id, p4_id],
              'old_port_pids': [],
              'former_port_pids': [p5_id, p4_id]}], [])
        # detach router from fifth and sixth subnets on same network, primary
        # must change since only one port remains
        del ports[1:3]
        # process again
        self.routing_helper._process_router(ri)
        # ----> VERIFY HERE: [.37]
        self._verify_operations(
            [],
            [{'p_id': p5_id,
              'new_port_pids': [],
              'current_port_pids': [p4_id],
              'old_port_pids': [p6_id, p5_id],
              'former_port_pids': [p6_id, p5_id, p4_id]},
             {'p_id': p6_id,
              'new_port_pids': [],
              'current_port_pids': [p4_id],
              'old_port_pids': [p6_id, p5_id],
              'former_port_pids': [p6_id, p5_id, p4_id]}])
        # detach router from fourth subnet on same network, primary can just
        # be deleted as there are no secondaries
        del ports[0]
        # process again
        self.routing_helper._process_router(ri)
        # ----> VERIFY HERE: None left
        self._verify_operations(
            [],
            [{'p_id': p4_id,
              'new_port_pids': [],
              'current_port_pids': [],
              'old_port_pids': [p4_id],
              'former_port_pids': [p4_id]}])

    # msn means that the router has an external gateway with multiple subnets
    def _test_process_msn_router(self, num_ext_subnets=3,
                                 test_admin_state=True):

        def _verify_ip_info(the_mock):
            self.assertEqual(num_ext_subnets, len(the_mock.ip_infos))
            self.assertTrue(the_mock.ip_infos[0]['is_primary'])
            self.assertEqual('19.4.0.1/28', the_mock.ip_infos[0]['ip_cidr'])
            for i in range(1, num_ext_subnets):
                self.assertFalse(the_mock.ip_infos[i]['is_primary'])
                self.assertEqual('19.4.%s.1/27' % i,
                                 the_mock.ip_infos[i]['ip_cidr'])

        # need these helpers to verify that ip_info is correct
        def _ext_gw_added(ri, ex_gw_port):
            self.routing_helper._external_gateway_added.ip_infos.append(
                copy.deepcopy(ex_gw_port['ip_info']))

        def _ext_gw_removed(ri, ex_gw_port):
            self.routing_helper._external_gateway_removed.ip_infos.append(
                copy.deepcopy(ex_gw_port['ip_info']))

        router, ports = self.prepare_router_data(
            num_ext_subnets=num_ext_subnets)
        # Setup mock for call to process floating ips
        self.routing_helper._process_router_floating_ips = mock.Mock()
        fake_floatingips1 = {'floatingips': [
            {'id': _uuid(),
             'floating_ip_address': '8.8.8.8',
             'fixed_ip_address': '7.7.7.7',
             'port_id': _uuid()}]}
        ri = routing_svc_helper.RouterInfo(router['id'], router=router)
        self.routing_helper._external_gateway_added.ip_infos = []
        self.routing_helper._external_gateway_added.side_effect = _ext_gw_added
        # Process with initial values
        self.routing_helper._process_router(ri)
        ex_gw_port = ri.router.get('gw_port')
        # Assert that process_floating_ips, internal_network & external network
        # added were all called with the right params
        (self.routing_helper._process_router_floating_ips
            .assert_called_once_with(ri, ex_gw_port))
        self.routing_helper._internal_network_added.assert_called_once_with(
            ri, ports[0], ex_gw_port)
        _verify_ip_info(self.routing_helper._external_gateway_added)
        self._reset_mocks()
        # remap floating IP to a new fixed ip
        fake_floatingips2 = copy.deepcopy(fake_floatingips1)
        fake_floatingips2['floatingips'][0]['fixed_ip_address'] = '7.7.7.8'
        router[bc.constants.FLOATINGIP_KEY] = fake_floatingips2['floatingips']

        # Process again and check that this time only the process_floating_ips
        # was only called.
        self.routing_helper._process_router(ri)
        ex_gw_port = ri.router.get('gw_port')
        (self.routing_helper._process_router_floating_ips
            .assert_called_once_with(ri, ex_gw_port))
        self.assertFalse(self.routing_helper._internal_network_added.called)
        self.assertFalse(self.routing_helper._external_gateway_added.called)
        self._reset_mocks()
        # remove just the floating ips
        del router[bc.constants.FLOATINGIP_KEY]
        # Process again and check that this time also only the
        # process_floating_ips and external_network remove was called
        self.routing_helper._process_router(ri)
        ex_gw_port = ri.router.get('gw_port')
        (self.routing_helper._process_router_floating_ips
            .assert_called_once_with(ri, ex_gw_port))
        self.assertFalse(self.routing_helper._internal_network_added.called)
        self.assertFalse(self.routing_helper._external_gateway_added.called)
        self._reset_mocks()

        if test_admin_state:
            self._test_router_admin_port_state(router, ri, ex_gw_port)

        # now no ports so state is torn down
        del router[bc.constants.INTERFACE_KEY]
        del router['gw_port']
        # Update router_info object
        ri.router = router
        # Keep a copy of the ex_gw_port before its gone after processing.
        ex_gw_port = ri.ex_gw_port
        self.routing_helper._external_gateway_removed.ip_infos = []
        self.routing_helper._external_gateway_removed.side_effect = (
            _ext_gw_removed)
        # Process router and verify that internal and external network removed
        # were called and floating_ips_process was called
        self.routing_helper._process_router(ri)
        self.assertFalse(self.routing_helper.
                         _process_router_floating_ips.called)
        self.assertFalse(self.routing_helper._external_gateway_added.called)
        self.routing_helper._internal_network_removed.assert_called_once_with(
            ri, ports[0], ex_gw_port)
        _verify_ip_info(self.routing_helper._external_gateway_removed)

    def test_process_router(self, test_admin_state=True):
        self._test_process_msn_router(num_ext_subnets=1,
                                      test_admin_state=test_admin_state)

    def test_process_msn_router(self):
        self._test_process_msn_router()

    def _test_process_global_msn_router(self, num_ext_subnets=3):

        # need these helpers to verify that ip_info is correct
        def _verify_ip_info(the_mock):
            self.assertEqual(num_ext_subnets, len(the_mock.ip_infos))
            self.assertTrue(the_mock.ip_infos[0]['is_primary'])
            self.assertEqual('35.4.0.4/24', the_mock.ip_infos[0]['ip_cidr'])
            for i in range(1, num_ext_subnets):
                self.assertFalse(the_mock.ip_infos[i]['is_primary'])
                self.assertEqual('35.4.%s.4/24' % i,
                                 the_mock.ip_infos[i]['ip_cidr'])

        def _int_nw_added(ri, port, ex_gw_port):
            self.assertIsNone(ex_gw_port)
            self.routing_helper._internal_network_added.ip_infos.append(
                copy.deepcopy(port['ip_info']))

        router, ports = self.prepare_router_data(
            num_ext_subnets=num_ext_subnets, is_global=True)
        # Setup mock for call to process floating ips
        self.routing_helper._process_router_floating_ips = mock.Mock()
        ri = routing_svc_helper.RouterInfo(router['id'], router=router)
        self.routing_helper._internal_network_added.ip_infos = []
        self.routing_helper._internal_network_added.side_effect = _int_nw_added
        # Process with initial values
        self.routing_helper._process_router(ri)
        self.assertFalse(self.routing_helper.
                         _process_router_floating_ips.called)
        self.assertFalse(self.routing_helper._external_gateway_added.called)
        _verify_ip_info(self.routing_helper._internal_network_added)
        self._reset_mocks()

        # now no ports so state is torn down
        del router[bc.constants.INTERFACE_KEY]
        # Update router_info object
        ri.router = router

        # Process router and verify that internal and external network removed
        # were called and floating_ips_process was called
        self.routing_helper._process_router(ri)
        self.assertFalse(self.routing_helper.
                         _process_router_floating_ips.called)
        self.assertFalse(self.routing_helper._external_gateway_added.called)
        self.assertFalse(self.routing_helper._external_gateway_removed.called)
        self.assertFalse(self.routing_helper._internal_network_added.called)
        self.routing_helper._internal_network_removed.assert_called_once_with(
            ri, ports[0], None)

    def test_process_global_router(self):
        self._test_process_global_msn_router(num_ext_subnets=1)

    def test_process_global_msn_router(self):
        self._test_process_global_msn_router()

    def test_routing_table_update(self):
        router = self.router
        fake_route1 = {'destination': '135.207.0.0/16',
                       'nexthop': '1.2.3.4'}
        fake_route2 = {'destination': '135.207.111.111/32',
                       'nexthop': '1.2.3.4'}

        # First we set the routes to fake_route1 and see if the
        # driver.routes_updated was called with 'replace'(==add or replace)
        # and fake_route1
        router['routes'] = [fake_route1]
        ri = routing_svc_helper.RouterInfo(router['id'], router)
        self.routing_helper._process_router(ri)

        self.driver.routes_updated.assert_called_with(ri, 'replace',
                                                      fake_route1)

        # Now we replace fake_route1 with fake_route2. This should cause driver
        # to be invoked to delete fake_route1 and 'replace'(==add or replace)
        self.driver.reset_mock()
        router['routes'] = [fake_route2]
        ri.router = router
        self.routing_helper._process_router(ri)

        self.driver.routes_updated.assert_called_with(ri, 'delete',
                                                      fake_route1)
        self.driver.routes_updated.assert_any_call(ri, 'replace', fake_route2)

        # Now we add back fake_route1 as a new route, this should cause driver
        # to be invoked to 'replace'(==add or replace) fake_route1
        self.driver.reset_mock()
        router['routes'] = [fake_route2, fake_route1]
        ri.router = router
        self.routing_helper._process_router(ri)

        self.driver.routes_updated.assert_any_call(ri, 'replace', fake_route1)

        # Now we delete all routes. This should cause driver
        # to be invoked to delete fake_route1 and fake-route2
        self.driver.reset_mock()
        router['routes'] = []
        ri.router = router
        self.routing_helper._process_router(ri)

        self.driver.routes_updated.assert_any_call(ri, 'delete', fake_route2)
        self.driver.routes_updated.assert_any_call(ri, 'delete', fake_route1)

    def test_process_router_internal_network_added_unexpected_error(self):
        router, ports = self.prepare_router_data()
        ri = routing_svc_helper.RouterInfo(router['id'], router=router)
        # raise RuntimeError to simulate that an unexpected exception occurrs
        self.routing_helper._internal_network_added.side_effect = RuntimeError
        self.assertRaises(RuntimeError,
                          self.routing_helper._process_router,
                          ri)
        self.assertNotIn(
            router[bc.constants.INTERFACE_KEY][0], ri.internal_ports)

        # The unexpected exception has been fixed manually
        self.routing_helper._internal_network_added.side_effect = None

        # Failure will cause a retry next time, then were able to add the
        # port to ri.internal_ports
        self.routing_helper._process_router(ri)
        self.assertIn(
            router[bc.constants.INTERFACE_KEY][0], ri.internal_ports)

    def test_process_router_internal_network_added_raises_HAMissingError(self):
        router, ports = self.prepare_router_data()
        router[ha.ENABLED] = True
        ri = routing_svc_helper.RouterInfo(router['id'], router=router)
        # raise RuntimeError to simulate that a HAParamsMissingException
        params = {'r_id': FAKE_ID, 'p_id': FAKE_ID, 'port': ports[0]}
        self.routing_helper._internal_network_added.side_effect = (
            cfg_exceptions.HAParamsMissingException(**params))
        self.routing_helper._process_router(ri)
        self.assertIn(ri.router_id, self.routing_helper.updated_routers)
        self.assertNotIn(
            router[bc.constants.INTERFACE_KEY][0], ri.internal_ports)
        # The unexpected exception has been fixed manually
        self.routing_helper._internal_network_added.side_effect = None

        # Failure will cause a retry next time, then were able to add the
        # port to ri.internal_ports
        self.routing_helper._process_router(ri)
        self.assertIn(
            router[bc.constants.INTERFACE_KEY][0], ri.internal_ports)

    def test_process_router_internal_network_removed_unexpected_error(self):
        router, ports = self.prepare_router_data()
        ri = routing_svc_helper.RouterInfo(router['id'], router=router)
        # add an internal port
        self.routing_helper._process_router(ri)

        # raise RuntimeError to simulate that an unexpected exception occurrs

        self.routing_helper._internal_network_removed.side_effect = mock.Mock(
            side_effect=RuntimeError)
        router[bc.constants.INTERFACE_KEY][0]['admin_state_up'] = False
        # The above port is set to down state, remove it.
        self.assertRaises(RuntimeError,
                          self.routing_helper._process_router,
                          ri)
        self.assertIn(
            router[bc.constants.INTERFACE_KEY][0], ri.internal_ports)

        # The unexpected exception has been fixed manually
        self.routing_helper._internal_network_removed.side_effect = None

        # Failure will cause a retry next time,
        # We were able to add the port to ri.internal_ports
        self.routing_helper._process_router(ri)
        # We were able to remove the port from ri.internal_ports
        self.assertNotIn(
            router[bc.constants.INTERFACE_KEY][0], ri.internal_ports)

    def test_routers_with_admin_state_down(self):
        self.plugin_api.get_external_network_id.return_value = None

        routers = [
            {'id': _uuid(),
             'admin_state_up': False,
             'external_gateway_info': {}}]
        self.routing_helper._process_routers(routers, None)
        self.assertNotIn(routers[0]['id'], self.routing_helper.router_info)

    def test_router_deleted(self):
        self.routing_helper.router_deleted(None, [FAKE_ID])
        self.assertIn(FAKE_ID, self.routing_helper.removed_routers)

    def test_routers_updated(self):
        self.routing_helper.routers_updated(None, [FAKE_ID])
        self.assertIn(FAKE_ID, self.routing_helper.updated_routers)

    def test_process_router_delete(self):
        router = self.router
        router['gw_port'] = self.ex_gw_port
        router[routerrole.ROUTER_ROLE_ATTR] = None
        self.routing_helper._router_added(router['id'], router)
        self.assertIn(router['id'], self.routing_helper.router_info)
        # Now we remove the router
        self.routing_helper._router_removed(router['id'], deconfigure=True)
        self.assertNotIn(router['id'], self.routing_helper.router_info)

    def test_collect_state(self):
        router, ports = self.prepare_router_data(enable_snat=True,
                                                 num_internal_ports=2)
        self.routing_helper._router_added(router['id'], router)

        configurations = {}
        configurations = self.routing_helper.collect_state(configurations)
        hd_exp_result = {
            router['hosting_device']['id']: {'routers': 1}}
        self.assertEqual(1, configurations['total routers'])
        self.assertEqual(1, configurations['total ex_gw_ports'])
        self.assertEqual(2, configurations['total interfaces'])
        self.assertEqual(0, configurations['total floating_ips'])
        self.assertEqual(hd_exp_result, configurations['hosting_devices'])
        self.assertEqual([], list(
            configurations['non_responding_hosting_devices']))

    def test_sort_resources_per_hosting_device(self):
        router1, port = self.prepare_router_data()
        router2, port = self.prepare_router_data()
        router3, port = self.prepare_router_data()
        router4, port = self.prepare_router_data()

        hd1_id = router1['hosting_device']['id']
        hd2_id = router4['hosting_device']['id']
        #Setting router2 and router3 device id same as router1's device id
        router2['hosting_device']['id'] = hd1_id
        router3['hosting_device']['id'] = hd1_id

        resources = {'routers': [router1, router2, router4],
                     'removed_routers': [router3]}
        devices = self.routing_helper._sort_resources_per_hosting_device(
            resources)

        self.assertEqual(2, len(devices.keys()))  # Two devices
        hd1_routers = [router1, router2]
        self.assertEqual(hd1_routers, devices[hd1_id]['routers'])
        self.assertEqual([router3], devices[hd1_id]['removed_routers'])
        self.assertEqual([router4], devices[hd2_id]['routers'])

    def test_get_router_ids_from_removed_devices_info(self):
        removed_devices_info = {
            'hosting_data': {'device_1': {'routers': ['id1', 'id2']},
                             'device_2': {'routers': ['id3', 'id4'],
                                          'other_key': ['value1', 'value2']}}
        }
        resp = self.routing_helper._get_router_ids_from_removed_devices_info(
            removed_devices_info)
        self.assertEqual(sorted(resp), sorted(['id1', 'id2', 'id3', 'id4']))

    @mock.patch("eventlet.GreenPool.spawn_n")
    def test_process_services_full_sync_different_devices(self, mock_spawn):
        router1, port = self.prepare_router_data()
        router2, port = self.prepare_router_data()
        self.plugin_api.get_routers = mock.Mock(
            return_value=[router1, router2])
        self.routing_helper.process_service()
        self.assertEqual(2, mock_spawn.call_count)
        call1 = mock.call(self.routing_helper._process_routers, [router1],
                          [], router1['hosting_device']['id'],
                          all_routers=True)
        call2 = mock.call(self.routing_helper._process_routers, [router2],
                          [], router2['hosting_device']['id'],
                          all_routers=True)
        mock_spawn.assert_has_calls([call1, call2], any_order=True)

    @mock.patch("eventlet.GreenPool.spawn_n")
    def test_process_services_full_sync_same_device(self, mock_spawn):
        router1, port = self.prepare_router_data()
        router2, port = self.prepare_router_data()
        router2['hosting_device']['id'] = router1['hosting_device']['id']
        self.plugin_api.get_routers = mock.Mock(return_value=[router1,
                                                              router2])
        self.routing_helper.process_service()
        self.assertEqual(1, mock_spawn.call_count)
        mock_spawn.assert_called_with(self.routing_helper._process_routers,
                                      [router1, router2], [],
                                      router1['hosting_device']['id'],
                                      all_routers=True)

    @mock.patch("eventlet.GreenPool.spawn_n")
    def test_process_services_with_updated_routers(self, mock_spawn):

        router1, port = self.prepare_router_data()

        def routers_data(context, router_ids=None, hd_ids=None):
            if router_ids:
                return [router1]
        self.plugin_api.get_routers.side_effect = routers_data

        self.routing_helper.fullsync = False
        self.routing_helper.updated_routers.add(router1['id'])
        self.routing_helper.process_service()
        self.assertEqual(1, self.plugin_api.get_routers.call_count)
        self.plugin_api.get_routers.assert_called_with(
            self.routing_helper.context,
            router_ids=[router1['id']])
        self.assertEqual(1, mock_spawn.call_count)
        mock_spawn.assert_called_with(self.routing_helper._process_routers,
                                      [router1], [],
                                      router1['hosting_device']['id'],
                                      all_routers=False)

    @mock.patch("eventlet.GreenPool.spawn_n")
    def test_process_services_with_deviceid(self, mock_spawn):

        router, port = self.prepare_router_data()
        device_id = router['hosting_device']['id']

        def routers_data(context, router_ids=None, hd_ids=None):
            if hd_ids:
                self.assertEqual([device_id], hd_ids)
                return [router]

        self.plugin_api.get_routers.side_effect = routers_data
        self.routing_helper.fullsync = False
        self.routing_helper.process_service(device_ids=[device_id])
        self.assertEqual(1, self.plugin_api.get_routers.call_count)
        self.plugin_api.get_routers.assert_called_with(
            self.routing_helper.context,
            hd_ids=[device_id])
        self.assertEqual(1, mock_spawn.call_count)
        mock_spawn.assert_called_with(self.routing_helper._process_routers,
                                      [router], [], device_id,
                                      all_routers=False)

    @mock.patch("eventlet.GreenPool.spawn_n")
    def test_process_services_with_removed_routers(self, mock_spawn):
        router, port = self.prepare_router_data()
        device_id = router['hosting_device']['id']

        self._mock_driver_and_hosting_device(self.routing_helper)
        self.routing_helper.fullsync = False
        # Emulate router added for setting up internal structures
        self.routing_helper._router_added(router['id'], router)
        # Add router to removed routers list and process it
        self.routing_helper.removed_routers.add(router['id'])
        self.routing_helper.process_service()

        self.assertEqual(1, mock_spawn.call_count)
        mock_spawn.assert_called_with(self.routing_helper._process_routers,
                                      [], [router], device_id,
                                      all_routers=False)

    @mock.patch("eventlet.GreenPool.spawn_n")
    def test_process_services_with_removed_routers_info(self, mock_spawn):
        router1, port = self.prepare_router_data()
        device_id = router1['hosting_device']['id']
        router2, port = self.prepare_router_data()
        router2['hosting_device']['id'] = _uuid()

        removed_devices_info = {
            'hosting_data': {device_id: {'routers': [router1['id']]}},
            'deconfigure': True
        }

        self._mock_driver_and_hosting_device(self.routing_helper)
        self.routing_helper.fullsync = False
        # Emulate router added for setting up internal structures
        self.routing_helper._router_added(router1['id'], router1)
        self.routing_helper._router_added(router2['id'], router2)
        # Add router to removed routers list and process it
        self.routing_helper.removed_routers.add(router2['id'])
        self.routing_helper.process_service(
            removed_devices_info=removed_devices_info)

        self.assertEqual(2, mock_spawn.call_count)
        call1 = mock.call(self.routing_helper._process_routers,
                          [], [router1], router1['hosting_device']['id'],
                          all_routers=False)
        call2 = mock.call(self.routing_helper._process_routers,
                          [], [router2], router2['hosting_device']['id'],
                          all_routers=False)
        mock_spawn.assert_has_calls([call1, call2], any_order=True)

    @mock.patch("eventlet.GreenPool.spawn_n")
    def test_process_services_with_rpc_error(self, mock_spawn):
        router, port = self.prepare_router_data()
        self.plugin_api.get_routers.side_effect = (
            oslo_messaging.MessagingException)
        self.routing_helper.fullsync = False
        self.routing_helper.updated_routers.add(router['id'])
        self.routing_helper.process_service()
        self.assertEqual(1, self.plugin_api.get_routers.call_count)
        self.plugin_api.get_routers.assert_called_with(
            self.routing_helper.context,
            router_ids=[router['id']])
        self.assertFalse(mock_spawn.called)
        self.assertTrue(self.routing_helper.fullsync)

    def test_process_routers(self):
        router, port = self.prepare_router_data()
        driver = self._mock_driver_and_hosting_device(self.routing_helper)
        self.routing_helper._process_router = mock.Mock()
        self.routing_helper._process_routers([router], None)
        ri = self.routing_helper.router_info[router['id']]
        driver.router_added.assert_called_with(ri)
        self.routing_helper._process_router.assert_called_with(ri)

    def _process_routers_floatingips(self, action='add'):
        router, port = self.prepare_router_data()
        driver = self._mock_driver_and_hosting_device(self.routing_helper)
        ex_gw_port = router['gw_port']
        floating_ip_address = '19.4.4.10'
        fixed_ip_address = '35.4.1.10'
        fixed_ip_address_2 = '35.4.1.15'
        port_id = 'fake_port_id'
        floating_ip = {'fixed_ip_address': fixed_ip_address,
                       'floating_ip_address': floating_ip_address,
                       'id': 'floating_ip_id',
                       'port_id': port_id,
                       'status': 'ACTIVE', }
        router[bc.constants.FLOATINGIP_KEY] = [floating_ip]
        ri = routing_svc_helper.RouterInfo(router['id'], router=router)

        # Default add action
        self.routing_helper._process_router_floating_ips(ri, ex_gw_port)
        driver.floating_ip_added.assert_called_with(
            ri, ex_gw_port, floating_ip_address, fixed_ip_address)

        if action == 'remove':
            router[bc.constants.FLOATINGIP_KEY] = []
            self.routing_helper._process_router_floating_ips(ri, ex_gw_port)
            driver.floating_ip_removed.assert_called_with(
                ri, ri.ex_gw_port, floating_ip_address, fixed_ip_address)

        if action == 'remap':
            driver.reset_mock()
            floating_ip_2 = copy.deepcopy(floating_ip)
            floating_ip_2['fixed_ip_address'] = fixed_ip_address_2
            ri.router[bc.constants.FLOATINGIP_KEY] = [floating_ip_2]

            self.routing_helper._process_router_floating_ips(ri, ex_gw_port)
            driver.floating_ip_added.assert_called_with(
                ri, ex_gw_port, floating_ip_address, fixed_ip_address_2)

            driver.floating_ip_removed.assert_called_with(
                ri, ri.ex_gw_port, floating_ip_address, fixed_ip_address)

    def test_process_routers_floatingips_add(self):
        self._process_routers_floatingips(action="add")

    def test_process_routers_floatingips_remove(self):
        self._process_routers_floatingips(action="remove")

    def test_process_routers_floatingips_remap(self):
        self._process_routers_floatingips(action="remap")

    def test_process_routers_rearrange_for_global(self):
        router1, port1 = self.prepare_router_data()

        router2 = {'id': _uuid(),
                   routerrole.ROUTER_ROLE_ATTR: None}
        router_G = {'id': _uuid(),
                   routerrole.ROUTER_ROLE_ATTR: c_constants.ROUTER_ROLE_GLOBAL}
        removed_routers = [router_G, router2]

        self.routing_helper._adjust_router_list_for_global_router(
                removed_routers)
        # we check if the routers were rearranged
        self.assertEqual(router2['id'], removed_routers[0]['id'])

        removed_routers = [router_G, router2]
        driver = self._mock_driver_and_hosting_device(self.routing_helper)
        self.routing_helper._process_router = mock.Mock()
        self.routing_helper._router_removed = mock.Mock()
        self.routing_helper._process_routers(
                [router1], removed_routers=removed_routers, device_id=None)
        ri1 = self.routing_helper.router_info[router1['id']]
        driver.router_added.assert_called_with(ri1)
        # This check ensures that the tenant router was deleted first, followed
        # by the global router
        self.routing_helper._router_removed.assert_called_with(router_G['id'])
        self.routing_helper._router_removed.assert_any_call(router2['id'])
        self.routing_helper._process_router.assert_called_with(ri1)

    def test_process_routers_skips_routers_on_other_hosting_devices(self):
        router1, port1 = self.prepare_router_data()
        r1_id = router1['id']
        r1_info = routing_svc_helper.RouterInfo(r1_id, router1)
        router2, port2 = self.prepare_router_data()
        r2_id = router2['id']
        self.routing_helper.router_info = {
            r1_id: r1_info,
            r2_id: routing_svc_helper.RouterInfo(r2_id, router2)}
        self.routing_helper._process_router = mock.Mock()
        self.routing_helper._router_removed = mock.Mock()
        self.routing_helper._process_routers([router1], [],
                                             router1['hosting_device']['id'])
        self.routing_helper._process_router.assert_called_once_with(r1_info)
        self.assertEqual(0, self.routing_helper._router_removed.call_count)

    def test_obfuscated_password_is_unobfuscated_for_driver(self):
        obfus_creds = dict(self.hosting_device['credentials'])
        a_cfg_agent = cfg_agent.CiscoCfgAgent('some_host')
        a_cfg_agent._credentials = {
            'creds_uuid1': {'credentials_id': 'creds_uuid1',
                            'username': 'user', 'password': '4getme'}}
        drv_mgr = driver_mgr.DeviceDriverManager(a_cfg_agent)
        driver = drv_mgr.set_driver(self.router)
        self.assertEqual(obfus_creds,
                         self.router['hosting_device']['credentials'])
        self.assertEqual('user', driver._username)
        self.assertEqual('4getme', driver._password)


class TestDeviceSyncOperations(base.BaseTestCase):

    def setUp(self):
        super(TestDeviceSyncOperations, self).setUp()
        self.conf = cfg.ConfigOpts()
        self.conf.register_opts(bc.core_opts)
        self.conf.register_opts(cfg_agent.OPTS, "cfg_agent")
        self.ex_gw_port = {'id': _uuid(),
                           'network_id': _uuid(),
                           'admin_state_up': True,
                           'fixed_ips': [{'ip_address': '19.4.4.4',
                                         'subnet_id': _uuid()}],
                           'subnets': [{'cidr': '19.4.4.0/24',
                                        'gateway_ip': '19.4.4.1'}]}
        self.hosting_device = {'id': "100",
                               'name': "VM_template",
                               'booting_time': 300,
                               'host_category': "VM",
                               'management_ip_address': '20.0.0.5',
                               'protocol_port': 22,
                               'credentials': {'credentials_id': 'creds_uuid1',
                                               'user_name': 'user',
                                               'password': '********'},
                               }

        self.fetched_routers = [
            {
                'id': _uuid(),
                'enable_snat': True,
                'admin_state_up': True,
                'routes': [],
                'gw_port': self.ex_gw_port,
                routerrole.ROUTER_ROLE_ATTR: None,
                'hosting_device': self.hosting_device
            },
            {
                'id': _uuid(),
                'enable_snat': True,
                'admin_state_up': True,
                'routes': [],
                'gw_port': self.ex_gw_port,
                routerrole.ROUTER_ROLE_ATTR: None,
                'hosting_device': self.hosting_device
            }

        ]

        #Patches & Mocks
        self.agent = mock.Mock()

        self.routing_helper = routing_svc_helper.RoutingServiceHelper(
            HOST, self.conf, self.agent)
        self.routing_helper._cleanup_invalid_cfg = mock.Mock()
        self.routing_helper._router_removed = mock.Mock()
        self.driver = self._mock_driver_and_hosting_device(
            self.routing_helper)

    def _mock_driver_and_hosting_device(self, svc_helper):
        svc_helper._dev_status.is_hosting_device_reachable = mock.MagicMock(
            return_value=True)
        driver = mock.MagicMock()
        svc_helper._drivermgr.get_driver = mock.Mock(return_value=driver)
        svc_helper._drivermgr.set_driver = mock.Mock(return_value=driver)
        return driver

    def _assert_called_once(self, mock_function):
        self.assertEqual(1, mock_function.call_count)

    def test_handle_sync_devices(self):
        self.routing_helper._fetch_router_info = (
            mock.Mock(return_value=self.fetched_routers))

        routers = []
        self.routing_helper._handle_sync_devices(routers)

        self._assert_called_once(self.routing_helper._fetch_router_info)
        self.assertEqual(2, self.routing_helper._router_removed.call_count)
        self._assert_called_once(self.routing_helper._cleanup_invalid_cfg)
        self.assertEqual(2, len(routers))

    def test_handle_sync_devices_retry(self):
        self.assertEqual(0, self.routing_helper.sync_devices_attempts)
        self.routing_helper._fetch_router_info = mock.Mock(return_value=None)

        self.routing_helper._handle_sync_devices([])

        self.assertEqual(1, self.routing_helper.sync_devices_attempts)

    def test_handle_sync_devices_exceed_max_retries(self):
        self.assertEqual(6, cfg.CONF.cfg_agent.max_device_sync_attempts)

        self.assertEqual(0, self.routing_helper.sync_devices_attempts)
        self.routing_helper._fetch_router_info = mock.Mock(return_value=None)

        for idx in range(1, 6):
            self.routing_helper._handle_sync_devices([])
            self.assertEqual(idx,
                             self.routing_helper.sync_devices_attempts)

        # expect sync_devices to be cleared and sync_devices_attempts == 0
        self.routing_helper._handle_sync_devices([])
        self.assertEqual(0,
                         self.routing_helper.sync_devices_attempts)
        self.assertEqual(0, len(self.routing_helper.sync_devices))

    def _test_sync_chunk_size(self, num_routers=10, exception_at=None,
                              current_chunk_size=None, max_chunk_size=None):
        def get_routers(context, router_ids):
            current_call_num = self.call_num
            self.call_num += 1
            if current_call_num in exception_at:
                raise oslo_messaging.MessagingTimeout
            return [routers[r_id] for r_id in router_ids if
                    r_id in routers]

        if max_chunk_size is not None:
            cfg.CONF.set_override('max_device_sync_batch_size', max_chunk_size,
                                  group='cfg_agent')
        if current_chunk_size is not None:
            self.routing_helper.sync_routers_chunk_size = current_chunk_size
        if exception_at is None:
            exception_at = []
        self.call_num = 0
        self.routing_helper.plugin_rpc = mock.MagicMock()
        r_ids = [_uuid() for i in range(num_routers)]
        routers = [{'id': uuid} for uuid in r_ids]
        self.routing_helper.plugin_rpc.get_router_ids = mock.MagicMock(
            return_value=r_ids)
        self.routing_helper.plugin_rpc.get_routers = mock.MagicMock(
            side_effect=get_routers)
        return r_ids, routers

    def test_message_timeout_reduces_sync_chunk_size(self):
        r_ids, routers = self._test_sync_chunk_size(exception_at=[0, 1, 2, 3])
        self.assertRaises(oslo_messaging.MessagingTimeout,
                          self.routing_helper._fetch_router_info, r_ids)
        self.assertEqual(32, self.routing_helper.sync_routers_chunk_size)
        self.assertRaises(oslo_messaging.MessagingTimeout,
                          self.routing_helper._fetch_router_info, r_ids)
        self.assertEqual(16, self.routing_helper.sync_routers_chunk_size)
        self.assertRaises(oslo_messaging.MessagingTimeout,
                          self.routing_helper._fetch_router_info, r_ids)
        self.assertEqual(8, self.routing_helper.sync_routers_chunk_size)
        self.assertRaises(oslo_messaging.MessagingTimeout,
                          self.routing_helper._fetch_router_info, r_ids)
        # should not go below the minimum chunk size
        self.assertEqual(8, self.routing_helper.sync_routers_chunk_size)

    def test_successful_fetch_increases_sync_chunk_size(self):
        r_ids, routers = self._test_sync_chunk_size(current_chunk_size=16,
                                                    max_chunk_size=28)
        self.routing_helper._fetch_router_info(r_ids)
        self.assertEqual(24, self.routing_helper.sync_routers_chunk_size)
        self.routing_helper._fetch_router_info(r_ids)
        # should not go above the maximum chunkk size
        self.assertEqual(28, self.routing_helper.sync_routers_chunk_size)



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import sys

import mock

from neutronclient import shell
from neutronclient.tests.unit import test_cli20

from networking_cisco.neutronclient import hostingdevicetemplate


class CLITestV20HostingDeviceTemplate(test_cli20.CLITestV20Base):

    def setUp(self):
        # need to mock before super because extensions loaded on instantiation
        self._mock_extension_loading()
        super(CLITestV20HostingDeviceTemplate, self).setUp()
        self.non_admin_status_resources.append('hosting_device_template')

    def _mock_extension_loading(self):
        ext_pkg = 'neutronclient.common.extension'
        contrib = mock.patch(ext_pkg + '._discover_via_entry_points').start()
        contrib.return_value = [("hostingdevicetemplate",
                                 hostingdevicetemplate)]
        return contrib

    def test_ext_cmd_loaded(self):
        shell.NeutronShell('2.0')
        ext_cmd = {
            'cisco-hosting-device-template-list':
                hostingdevicetemplate.HostingDeviceTemplateList,
            'cisco-hosting-device-template-create':
                hostingdevicetemplate.HostingDeviceTemplateCreate,
            'cisco-hosting-device-template-update':
                hostingdevicetemplate.HostingDeviceTemplateUpdate,
            'cisco-hosting-device-template-delete':
                hostingdevicetemplate.HostingDeviceTemplateDelete,
            'cisco-hosting-device-template-show':
                hostingdevicetemplate.HostingDeviceTemplateShow}
        self.assertDictContainsSubset(ext_cmd, shell.COMMANDS['2.0'])

    def test_ext_cmd_help_doc_with_extension_name(self):
        shell.NeutronShell('2.0')
        ext_cmd = {
            'cisco-hosting-device-template-list':
                hostingdevicetemplate.HostingDeviceTemplateList,
            'cisco-hosting-device-template-create':
                hostingdevicetemplate.HostingDeviceTemplateCreate,
            'cisco-hosting-device-template-update':
                hostingdevicetemplate.HostingDeviceTemplateUpdate,
            'cisco-hosting-device-template-delete':
                hostingdevicetemplate.HostingDeviceTemplateDelete,
            'cisco-hosting-device-template-show':
                hostingdevicetemplate.HostingDeviceTemplateShow}
        self.assertDictContainsSubset(ext_cmd, shell.COMMANDS['2.0'])
        for item in ext_cmd:
            cmdcls = shell.COMMANDS['2.0'].get(item)
            self.assertTrue(cmdcls.__doc__.startswith(
                "[hostingdevicetemplate]"))

    def test_create_hosting_device_template(self):
        """Create hosting device template."""
        resource = 'hosting_device_template'
        cmd = hostingdevicetemplate.HostingDeviceTemplateCreate(
            test_cli20.MyApp(sys.stdout), None)
        name = 'Device template 1'
        host_category = 'Hardware'
        myid = 'myid'
        args = [name, host_category]
        position_names = ['name', 'host_category']
        position_values = [name, host_category]
        self._test_create_resource(resource, cmd, None, myid, args,
                                   position_names, position_values)

    def test_create_hosting_device_template_id(self):
        """Create hosting device template: --id this_id "Device 1" "Template
        1".
        """
        resource = 'hosting_device_template'
        cmd = hostingdevicetemplate.HostingDeviceTemplateCreate(
            test_cli20.MyApp(sys.stdout), None)
        name = 'Device template 1'
        host_category = 'Hardware'
        myid = 'myid'
        args = ['--id', myid, name, host_category]
        position_names = ['name', 'host_category']
        position_values = [name, host_category]
        self._test_create_resource(resource, cmd, None, myid, args,
                                   position_names, position_values,
                                   id=myid)

    def test_create_hosting_device_template_tenant(self):
        """Create hosting device template: --tenant_id tenantid "Device 1"
        "Template 1".
        """
        resource = 'hosting_device_template'
        cmd = hostingdevicetemplate.HostingDeviceTemplateCreate(
            test_cli20.MyApp(sys.stdout), None)
        name = 'Device template 1'
        host_category = 'Hardware'
        myid = 'myid'
        args = ['--tenant_id', 'tenantid', name, host_category]
        position_names = ['name', 'host_category']
        position_values = [name, host_category]
        self._test_create_resource(resource, cmd, None, myid, args,
                                   position_names, position_values,
                                   tenant_id='tenantid')

    def _test_create_hosting_device_template_optional_args(
            self, disabled=None, service_types=None, servicetypes=None,
            image=None, flavor=None, creds_id=None, credsid=None,
            conf_mech=None, confmech=None, proto_port=None, protoport=None,
            boot_time=None, boottime=None, slot_capacity=None,
            slotcapacity=None, slots_free=None, slotsfree=None,
            tenant_bound=None, tenantbound=None, dev_drv=None, devdrv=None,
            plug_drv=None, plugdrv=None):
        resource = 'hosting_device_template'
        cmd = hostingdevicetemplate.HostingDeviceTemplateCreate(
            test_cli20.MyApp(sys.stdout), None)
        myid = 'myid'
        name = 'Device template 1'
        host_category = 'Hardware'
        args = []
        expected = {}
        if disabled is not None:
            args += ['--disabled']
            expected['enabled'] = False
        if service_types is not None:
            args += ['--service_types', service_types]
            expected['service_types'] = service_types
        if servicetypes is not None:
            args += ['--service-types', servicetypes]
            expected['service_types'] = servicetypes
        if image is not None:
            args += ['--image', image]
            expected['image'] = image
        if flavor is not None:
            args += ['--flavor', flavor]
            expected['flavor'] = flavor
        if creds_id is not None:
            args += ['--credentials_id', creds_id]
            expected['default_credentials_id'] = creds_id
        if credsid is not None:
            args += ['--credentials-id', credsid]
            expected['default_credentials_id'] = credsid
        if conf_mech is not None:
            args += ['--config_mechanism', conf_mech]
            expected['configuration_mechanism'] = conf_mech
        if confmech is not None:
            args += ['--config-mechanism', confmech]
            expected['configuration_mechanism'] = confmech
        if proto_port is not None:
            args += ['--protocol_port', proto_port]
            expected['protocol_port'] = proto_port
        if protoport is not None:
            args += ['--protocol-port', protoport]
            expected['protocol_port'] = protoport
        if boot_time is not None:
            args += ['--booting_time', boot_time]
            expected['booting_time'] = boot_time
        if boottime is not None:
            args += ['--booting-time', boottime]
            expected['booting_time'] = boottime
        if slot_capacity is not None:
            args += ['--slot_capacity', slot_capacity]
            expected['slot_capacity'] = slot_capacity
        if slotcapacity is not None:
            args += ['--slot-capacity', slotcapacity]
            expected['slot_capacity'] = slotcapacity
        if slots_free is not None:
            args += ['--desired_slots_free', slots_free]
            expected['desired_slots_free'] = slots_free
        if slotsfree is not None:
            args += ['--desired-slots-free', slotsfree]
            expected['desired_slots_free'] = slotsfree
        if dev_drv is not None:
            args += ['--device_driver', dev_drv]
            expected['device_driver'] = dev_drv
        if devdrv is not None:
            args += ['--device-driver', devdrv]
            expected['device_driver'] = devdrv
        if plug_drv is not None:
            args += ['--plugging_driver', plug_drv]
            expected['plugging_driver'] = plug_drv
        if plugdrv is not None:
            args += ['--plugging-driver', plugdrv]
            expected['plugging_driver'] = plugdrv
        if tenant_bound is not None:
            args += ['--tenant_bound', tenant_bound]
            expected['tenant_bound'] = (tenant_bound
                                        if tenant_bound != "None" else None)
        if tenantbound is not None:
            args += ['--tenant-bound', tenantbound]
            expected['tenant_bound'] = (tenantbound
                                        if tenantbound != "None" else None)
        position_names = ['name', 'host_category']
        position_values = [name, host_category]
        for p_v in position_values:
            args.append(p_v)
        self._test_create_resource(resource, cmd, None, myid, args,
                                   position_names, position_values,
                                   **expected)

    def test_create_hosting_device_template_disabled(self):
        self._test_create_hosting_device_template_optional_args(disabled=True)

    def test_create_hosting_device_template_service_types(self):
        self._test_create_hosting_device_template_optional_args(
            service_types="L3,FW")
        self._test_create_hosting_device_template_optional_args(
            servicetypes="L3,FW")

    def test_create_hosting_device_template_image(self):
        self._test_create_hosting_device_template_optional_args(
            image="some_glance_image")

    def test_create_hosting_device_template_flavor(self):
        self._test_create_hosting_device_template_optional_args(
            flavor="some_nova_flavor")

    def test_create_hosting_device_template_creds(self):
        self._test_create_hosting_device_template_optional_args(
            creds_id='some id')
        self._test_create_hosting_device_template_optional_args(
            credsid='some id')

    def test_create_hosting_device_template_conf_mech(self):
        self._test_create_hosting_device_template_optional_args(
            conf_mech='netconf')
        self._test_create_hosting_device_template_optional_args(
            conf_mech='netconf')

    def test_create_hosting_device_template_proto_port(self):
        self._test_create_hosting_device_template_optional_args(
            proto_port='22')
        self._test_create_hosting_device_template_optional_args(
            protoport='22')

    def test_create_hosting_device_template_boot_time(self):
        self._test_create_hosting_device_template_optional_args(
            boot_time='3000')
        self._test_create_hosting_device_template_optional_args(
            boottime='3000')

    def test_create_hosting_device_template_slot_capacity(self):
        self._test_create_hosting_device_template_optional_args(
            slot_capacity='10000')
        self._test_create_hosting_device_template_optional_args(
            slotcapacity='10000')

    def test_create_hosting_device_template_desired_slots_free(self):
        self._test_create_hosting_device_template_optional_args(
            slots_free='200')
        self._test_create_hosting_device_template_optional_args(
            slotsfree='200')

    def test_create_hosting_device_template_tenant_bound(self):
        self._test_create_hosting_device_template_optional_args(
            tenant_bound='None')
        self._test_create_hosting_device_template_optional_args(
            tenantbound='None')
        self._test_create_hosting_device_template_optional_args(
            tenant_bound='')
        self._test_create_hosting_device_template_optional_args(
            tenantbound='')
        self._test_create_hosting_device_template_optional_args(
            tenant_bound='some id')
        self._test_create_hosting_device_template_optional_args(
            tenantbound='some id')

    def test_create_hosting_device_template_device_driver(self):
        self._test_create_hosting_device_template_optional_args(
            dev_drv='my.device.driver:class_name')
        self._test_create_hosting_device_template_optional_args(
            devdrv='my.device.driver:class_name')

    def test_create_hosting_device_template_plugging_driver(self):
        self._test_create_hosting_device_template_optional_args(
            plug_drv='my.plugging.driver:class_name')
        self._test_create_hosting_device_template_optional_args(
            plugdrv='my.plugging.driver:class_name')

    def test_create_hosting_device_template_full(self):
        self._test_create_hosting_device_template_optional_args(
            disabled=True, service_types="L3,FW", image="some_glance_image",
            flavor="some_nova_flavor", creds_id="some id",
            conf_mech="netconf", proto_port="22", boot_time="3000",
            slot_capacity="20000", slots_free="150", tenant_bound="None",
            dev_drv='my.device.driver:class_name',
            plug_drv='my.plugging.driver:class_name')
        self._test_create_hosting_device_template_optional_args(
            disabled=True, servicetypes="L3,FW", image="some_glance_image",
            flavor="some_nova_flavor", credsid="some id",
            conf_mech="netconf", proto_port="22", boottime="3000",
            slotcapacity="20000", slotsfree="150", tenantbound="None",
            devdrv='my.device.driver:class_name',
            plugdrv='my.plugging.driver:class_name')

    def test_list_hosting_device_templates_detail(self):
        """list hosting device templates: -D."""
        resources = "hosting_device_templates"
        cmd = hostingdevicetemplate.HostingDeviceTemplateList(
            test_cli20.MyApp(sys.stdout), None)
        response_contents = [{
            self.id_field: 'myid1', 'name': 'device_template_1_name',
            'host_category': 'VM', 'service_types': "L3,FW",
            'image': 'some_glance_image_1', 'enabled': True}, {
            self.id_field: 'myid2', 'name': 'device_template_2_name',
            'host_category': 'VM', 'service_types': "L3,FW",
            'image': 'some_glance_image_2', 'enabled': True}]
        self._test_list_resources(resources, cmd, True,
                                  response_contents=response_contents)

    def test_list_hosting_device_templates_sort(self):
        """list hosting device templates: --sort-key name --sort-key id
        --sort-key asc
        --sort-key desc
        """
        resources = "hosting_device_templates"
        cmd = hostingdevicetemplate.HostingDeviceTemplateList(
            test_cli20.MyApp(sys.stdout), None)
        response_contents = [{
            self.id_field: 'myid1', 'name': 'device_template_1_name',
            'host_category': 'VM', 'service_types': "L3,FW",
            'image': 'some_glance_image_1', 'enabled': True}, {
            self.id_field: 'myid2', 'name': 'device_template_2_name',
            'host_category': 'VM', 'service_types': "L3,FW",
            'image': 'some_glance_image_2', 'enabled': True}]
        self._test_list_resources(resources, cmd,
                                  sort_key=["name", "id"],
                                  sort_dir=["asc", "desc"],
                                  response_contents=response_contents)

    def test_list_hosting_device_templates_limit(self):
        """list hosting device templates: -P."""
        resources = "hosting_device_templates"
        cmd = hostingdevicetemplate.HostingDeviceTemplateList(
            test_cli20.MyApp(sys.stdout), None)
        response_contents = [{
            self.id_field: 'myid1', 'name': 'device_template_1_name',
            'host_category': 'VM', 'service_types': "L3,FW",
            'image': 'some_glance_image_1', 'enabled': True}, {
            self.id_field: 'myid2', 'name': 'device_template_2_name',
            'host_category': 'VM', 'service_types': "L3,FW",
            'image': 'some_glance_image_2', 'enabled': True}]
        self._test_list_resources(resources, cmd, page_size=1000,
                                  response_contents=response_contents)

    def _test_update_hosting_device_template(
            self, name=None, disabled=None, service_types=None,
            servicetypes=None, image=None, flavor=None, creds_id=None,
            credsid=None, conf_mech=None, confmech=None, proto_port=None,
            protoport=None, boot_time=None, boottime=None, tenant_bound=None,
            tenantbound=None):
        resource = 'hosting_device_template'
        cmd = hostingdevicetemplate.HostingDeviceTemplateUpdate(
            test_cli20.MyApp(sys.stdout), None)
        myid = 'myid'
        args = [myid]
        expected = {}
        if name is not None:
            args += ['--name', name]
            expected['name'] = name
        if disabled is not None:
            args += ['--disabled']
            expected['enabled'] = False
        if service_types is not None:
            args += ['--service_types', service_types]
            expected['service_types'] = service_types
        if servicetypes is not None:
            args += ['--service-types', servicetypes]
            expected['service_types'] = servicetypes
        if image is not None:
            args += ['--image', image]
            expected['image'] = image
        if flavor is not None:
            args += ['--flavor', flavor]
            expected['flavor'] = flavor
        if creds_id is not None:
            args += ['--credentials_id', creds_id]
            expected['default_credentials_id'] = creds_id
        if credsid is not None:
            args += ['--credentials-id', credsid]
            expected['default_credentials_id'] = credsid
        if conf_mech is not None:
            args += ['--config_mechanism', conf_mech]
            expected['configuration_mechanism'] = conf_mech
        if confmech is not None:
            args += ['--config-mechanism', confmech]
            expected['configuration_mechanism'] = confmech
        if proto_port is not None:
            args += ['--protocol_port', proto_port]
            expected['protocol_port'] = proto_port
        if protoport is not None:
            args += ['--protocol-port', protoport]
            expected['protocol_port'] = protoport
        if boot_time is not None:
            args += ['--booting_time', boot_time]
            expected['booting_time'] = boot_time
        if boottime is not None:
            args += ['--booting-time', boottime]
            expected['booting_time'] = boottime
        if tenant_bound is not None:
            args += ['--tenant_bound', tenant_bound]
            expected['tenant_bound'] = (tenant_bound
                                        if tenant_bound != "None" else None)
        if tenantbound is not None:
            args += ['--tenant-bound', tenantbound]
            expected['tenant_bound'] = (tenantbound
                                        if tenantbound != "None" else None)
        self._test_update_resource(resource, cmd, myid, args, expected)

    def test_update_hosting_device_template_disabled(self):
        self._test_update_hosting_device_template(disabled=True)

    def test_update_hosting_device_template_service_types(self):
        self._test_update_hosting_device_template(service_types="L3,FW")
        self._test_update_hosting_device_template(servicetypes="L3,FW")

    def test_update_hosting_device_template_image(self):
        self._test_update_hosting_device_template(image="some_glance_image")

    def test_update_hosting_device_template_flavor(self):
        self._test_update_hosting_device_template(flavor="some_nova_flavor")

    def test_update_hosting_device_template_creds(self):
        self._test_update_hosting_device_template(creds_id='some id')
        self._test_update_hosting_device_template(credsid='some id')

    def test_update_hosting_device_template_conf_mech(self):
        self._test_update_hosting_device_template(conf_mech='netconf')
        self._test_update_hosting_device_template(conf_mech='netconf')

    def test_update_hosting_device_template_proto_port(self):
        self._test_update_hosting_device_template(proto_port='22')
        self._test_update_hosting_device_template(protoport='22')

    def test_update_hosting_device_template_boot_time(self):
        self._test_update_hosting_device_template(boot_time='3000')
        self._test_update_hosting_device_template(boottime='3000')

    def test_update_hosting_device_template_tenant_bound(self):
        self._test_update_hosting_device_template(tenant_bound='None')
        self._test_update_hosting_device_template(tenantbound='None')
        self._test_update_hosting_device_template(tenant_bound='')
        self._test_update_hosting_device_template(tenantbound='')
        self._test_update_hosting_device_template(tenant_bound='some id')
        self._test_update_hosting_device_template(tenantbound='some id')

    def test_update_hosting_device_template_full(self):
        self._test_update_hosting_device_template(
            disabled=True, service_types="L3,FW", image="some_glance_image",
            flavor="some_nova_flavor", creds_id="some id",
            conf_mech="netconf", proto_port="22", boot_time="3000",
            tenant_bound="None")
        self._test_update_hosting_device_template(
            disabled=True, servicetypes="L3,FW", image="some_glance_image",
            flavor="some_nova_flavor", credsid="some id",
            conf_mech="netconf", proto_port="22", boottime="3000",
            tenantbound="None")

    def test_delete_hosting_device_template(self):
        """Delete hosting device template: myid."""
        resource = 'hosting_device_template'
        cmd = hostingdevicetemplate.HostingDeviceTemplateDelete(
            test_cli20.MyApp(sys.stdout), None)
        myid = 'myid'
        args = [myid]
        self._test_delete_resource(resource, cmd, myid, args)

    def test_show_hosting_device_template(self):
        """Show hosting device template: myid."""
        resource = 'hosting_device_template'
        cmd = hostingdevicetemplate.HostingDeviceTemplateShow(
            test_cli20.MyApp(sys.stdout), None)
        args = ['--fields', 'id', '--fields', 'name', self.test_id]
        self._test_show_resource(resource, cmd, self.test_id, args,
                                 ['id', 'name'])



# Copyright 2017 Cisco Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""nexus_provider_network

Revision ID: 972479e0e629
Revises: f3765e42de23
Create Date: 2017-05-02 15:11:15.280677

"""

# revision identifiers, used by Alembic.
revision = '972479e0e629'
down_revision = 'f3765e42de23'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('cisco_ml2_nexus_provider_networks',
        sa.Column('network_id', sa.String(length=36), nullable=False),
        sa.Column('vlan_id', sa.Integer, nullable=False, index=True),
        sa.PrimaryKeyConstraint('network_id'),
    )

    op.drop_column('cisco_ml2_nexusport_bindings', 'is_provider_vlan')
